/**
 * This file contains functions specific to the drop viewer ( /imagingsession/NNN ).
 */

/**********************************************************************************************************
 * KEYBOARD SHORTCUTS HANDLER
 * Note that the actual key mappings are defined in dropviewer_keymappings.js
 **********************************************************************************************************/

let KeyboardShortcuts={

		ctrlPressed:false,
		detectCtrlKey:function(evt){
			if((evt.keyCode && evt.keyCode===17) || (evt.key && "Control"===evt.key)){
				if(evt.type==="keydown"){
					KeyboardShortcuts.ctrlPressed=true;
				} else if(evt.type==="keyup"){
					KeyboardShortcuts.ctrlPressed=false;
				}
			}
		},
		
		shiftPressed:false,
		detectShiftKey:function(evt){
			if((evt.keyCode && evt.keyCode===16) || (evt.key && "Shift"===evt.key)){
				if(evt.type==="keydown"){
					KeyboardShortcuts.shiftPressed=true;
				} else if(evt.type==="keyup"){
					KeyboardShortcuts.shiftPressed=false;
				}
			}
		},
		
		keyMapping:new Map([]),
		
		init:function(){
			let mapping=UserConfig.get("dropviewer_keyboardMapping","British");
			KeyboardShortcuts.keyMapping=KeyboardShortcuts.keyMappings[mapping];
			document.observe("keydown", KeyboardShortcuts.handleKeyDown);
			document.observe("keyup", KeyboardShortcuts.handleKeyUp);
			document.observe("keydown", KeyboardShortcuts.detectCtrlKey);
			document.observe("keyup", KeyboardShortcuts.detectCtrlKey);
			document.observe("keydown", KeyboardShortcuts.detectShiftKey);
			document.observe("keyup", KeyboardShortcuts.detectShiftKey);
		},

		handleKeyUp:function(evt){
			KeyboardShortcuts.handleKey(evt,false);
		},
		handleKeyDown:function(evt) {
			KeyboardShortcuts.handleKey(evt,true);
		},
		handleKey:function(evt, isDown){
			if(document.activeElement){
				let type=document.activeElement.tagName.toLowerCase();
				if("input"===type||"textarea"===type||"select"===type){ return false;}
			}
			let keyPressed;
			let keyHandler;
			if(evt.key){
				keyPressed=evt.key;
			}
			keyHandler=KeyboardShortcuts.keyMapping[keyPressed];

			if(isDown) {
				if(keyHandler){
					keyHandler(keyPressed);
				}
			} else {
				//call the mirroring function - swap out images swapped in on keydown, for example
				if (keyHandler === DropViewer.swapInTimeCourseFirstImage) {
					DropViewer.swapOutTimeCourseFirstImage();
				} else if (keyHandler === DropViewer.swapInTimeCourseLastImage) {
					DropViewer.swapOutTimeCourseLastImage();
				}
			}
		},

		movieFirst:function(){
			if(DropNav.isOpen){
				DropNav.goToFirst();
			} else if(currentPanel.goToFirst){
				currentPanel.goToFirst();
			} else {
				DropViewer.goToFirst();
			}
		},
		moviePrevious:function(){
			if(DropNav.isOpen){
				DropNav.goToPrevious();
			} else if(currentPanel.goToPrevious){
				currentPanel.goToPrevious();
			} else {
				DropViewer.goToPrevious(true);
			}
		},
		movieNext:function(){
			if(DropNav.isOpen){
				DropNav.goToNext();
			} else if(currentPanel.goToNext){
				currentPanel.goToNext();
			} else {
				DropViewer.goToNext(true);
			}
		},
		movieLast:function(){
			if(DropNav.isOpen){
				DropNav.goToLast();
			} else if(currentPanel.goToLast){
				currentPanel.goToLast();
			} else {
				DropViewer.goToLast();
			}
		},
		moviePlayStop:function(){
			if(DropNav.isOpen){
				return false;
			} else if(currentPanel.playOrStop){
				currentPanel.playOrStop();
			} else {
				DropViewer.playOrStop();
			}
		},

		upOneRow:function(){
			if(DropNav.isOpen){
				DropNav.goUpOneRow();
			} else if(currentPanel.goUpOneRow){
				currentPanel.goUpOneRow();
			} else {
				console.log("No goUpOneRow in "+currentPanel.name)
			}
		},
		downOneRow:function(){
			if(DropNav.isOpen){
				DropNav.goDownOneRow();
			} else if(currentPanel.goDownOneRow){
				currentPanel.goDownOneRow();
			} else {
				console.log("No goUpOneRow in "+currentPanel.name)
			}
		},
		
		toggleTimeCourse:function(){
			if(currentPanel===TimeCourse){
				switchPanels($("viewbutton_dv"));
			} else {
				switchPanels($("viewbutton_tc"));
			}
		},
		
		olderImage:function(){
			if(currentPanel.olderImage){
				currentPanel.olderImage();
			}
		},
		
		newerImage:function(){
			if(currentPanel.newerImage){
				currentPanel.newerImage();
			}
		},
		
};



/**********************************************************************************************************
 * DROP VIEWER
 **********************************************************************************************************/

window.DropViewer={

	//array is zero-based, plate row numbers are 1-based, so empty item first.
	rowLabels:['','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T'],

	movieDelay:1.0, //seconds
	movieInterval:null,
	plateType:null,
	screen:null,
	currentIndex:0,
	maxIndex:0,
	images:[],

	warningImage:'<img alt="" style="height:1.5em;vertical-align:top;position:relative;top:-0.1em;margin-right:0.25em" src="/images/icons/'+skin.bodyicontheme+'/warning.png" />',

	init:function(){
		document.title="Drop viewer: Plate "+data['platename'];
		setHandedness(UserConfig.get("dropviewer_handedness","right"));
		DropViewer.movieDelay=UserConfig.get("dropviewer_movieDelay", DropViewer.movieDelay);
		$("dv_delay").innerHTML=DropViewer.movieDelay;
		$("tc_delay").innerHTML=DropViewer.movieDelay;
		DropViewer.getScoringSystem();
		DropViewer.getPlateType();
		DropViewer.getScreen();
		new Ajax.Request('/api/imagingsession/'+data['id']+'/dropimage?pagesize=10000',{
			method:'get',
			onSuccess:DropViewer.init_onSuccess,
			onFailure:DropViewer.init_onFailure
		});
		Event.observe(window, "resize", DropViewer.fitImageToPane);
	},

	init_onSuccess:function(transport){
		let images=transport.responseJSON.rows;
		let img=$("currentimage");
		img.src=images[0].fullimageurl;
		img.dataset.micronshigh=images[0].micronsperpixely*images[0].pixelheight+"";
		DropViewer.images=images;
		DropViewer.maxIndex=images.length-1;

		//Determine starting state from hash - drop and, if present, crystal. Example: #A03.2 #H7.2c4
		//Otherwise from user config, defaulting to measure/score
		let startImage=0;
		let startCrystal=-1;
		if(document.location.hash){
			let dropLabel=document.location.hash.substring(1);
			let startRow=DropViewer.rowLabels.indexOf(dropLabel.substr(0,1));
			let parts=dropLabel.substring(1).split(".");
			let startColumn=parseInt(parts[0]);
			let startDrop=parts[1];
			if(-1!==startDrop.indexOf("c")){
				let subParts=startDrop.split("c");
				startDrop=parseInt(subParts[0]);
				startCrystal=parseInt(subParts[1]);
				window.startMode="CrystalSelection";
			}
			let numImages=DropViewer.images.length;
			for(let i=0;i<numImages;i++){
				let img=DropViewer.images[i];
				let imgRow=parseInt(img.row);
				let imgCol=parseInt(img.col);
				let imgDrop=parseInt(img.dropnumber);
				if(startRow===imgRow && startColumn===imgCol && parseInt(startDrop)===imgDrop){
					startImage=i;
					break;
				}
			}
		}

		DropViewer.getCurrentImageTimeCourse();
		window.setTimeout(DropViewer.highlightCurrentImageScore,250);
		DropViewer.setCurrentImage(startImage);

		let startMode=UserConfig.get("dropviewer_startMode","DropViewer");
		currentPanel=startMode;
		window.setTimeout(function(){
			if(window[startMode]){
				window[startMode].switchTo();
			} else {
				DropViewer.switchTo();
				UserConfig.set("dropviewer_startMode","DropViewer");
			}
		},250);

		if(startCrystal>0){
			CrystalSelection.currentCrystalNumber=startCrystal;
			window.setTimeout(CrystalSelection.switchTo, 250);
		}
		window.setTimeout(Measure.setScale,250);
		
		window.setTimeout(DropViewer.preCacheMovieImages,50);
	},

	init_onFailure:function(){
		alert("Could not get images.");
	},

	getScreen:function(){
		if(""===data["screenid"] || null==data["screenid"] || 0===data["screenid"]){ return false; }
		new Ajax.Request('/api/screen/'+data["screenid"],{
			method:'get',
			onSuccess:DropViewer.getScreen_onSuccess,
			onFailure:DropViewer.getScreen_onFailure
		});
	},
	getScreen_onSuccess:function(transport){
		AjaxUtils.checkResponse(transport);
		window.setTimeout(function(){
			DropViewer.screen=transport.responseJSON;
			new Ajax.Request('/api/screen/'+data["screenid"]+'/screencondition?pagesize=10000&sortby=name',{
				method:'get',
				onSuccess:function(transport){ 
					AjaxUtils.checkResponse(transport);
					DropViewer.screen.conditions=transport.responseJSON.rows;
				},
				onFailure:function(){
					/* do nothing, assume screen not set */
				}
			});
		},50);
	},
	getScreen_onFailure:function(){
		/* do nothing, assume screen not set */
	},

	getPlateType:function(){
		new Ajax.Request('/api/platetype/'+data["platetypeid"],{
			method:'get',
			onSuccess:DropViewer.getPlateType_onSuccess,
			onFailure:DropViewer.getPlateType_onFailure
		});
	},
	getPlateType_onSuccess:function(transport){
		AjaxUtils.checkResponse(transport);
		window.setTimeout(function(){
			DropViewer.plateType=transport.responseJSON;
		},50);
	},
	getPlateType_onFailure:function(){
		alert("Could not retrieve plate type information");
	},
	
	
	getScoringSystem:function(){
		new Ajax.Request('/api/crystalscoringsystem/'+data["crystalscoringsystemid"]+'/score?sortby=scoreindex',{
			method:'get',
			onSuccess:DropViewer.getScoringSystem_onSuccess,
			onFailure:DropViewer.getScoringSystem_onFailure
		});
	},
	getScoringSystem_onSuccess:function(transport){
		DropViewer.scores=transport.responseJSON.rows;
		AjaxUtils.checkResponse(transport);
		if(50>$("dv_scores").getHeight()){
			window.setTimeout(function(){DropViewer.getScoringSystem_onSuccess(transport)},50);
			return;
		}
		DropViewer.renderScoringSystem(transport);
	},
	
	/* If the score buttons would be below this height, render them double-height half-width.
	 * Intent: Make them more touchable on small tablets.
	 */
	scoreHeightCutoff:50,
	
	renderScoringSystem:function(transport){
		let scores=transport.responseJSON.rows;
		DropViewer.scores=scores;
		DropViewer.scoreIndicesByScoreId={};
		let total=transport.responseJSON.rows.length;
		let scoreHeight=Math.floor(  $("dv_scores").getHeight() / (1+Math.round(total)) );
		let renderHalfWidth=false;
		if(scoreHeight<DropViewer.scoreHeightCutoff){
			renderHalfWidth=true;
		}
		$("dv_scores").innerHTML="";
		scores.each(function(s){
			DropViewer.scoreIndicesByScoreId['s'+s.id]=1*s.scoreindex;
			let sc=document.createElement("div");
			sc.id="score"+s.id;
			sc.dataset.scoreid=s.id;
			sc.dataset.scoreindex=s.scoreindex;
			sc.dataset.hotkey=s.hotkey;
			sc.className="dv_score";
			sc.style.height=(scoreHeight-4)+"px";
			sc.style.backgroundColor="black";
			sc.style.borderColor="#"+s.color;
			sc.style.fontWeight="bold";
			sc.style.paddingLeft="1em";
			sc.style.textShadow="1px 1px 1px black";
			sc.innerHTML='['+s.hotkey+']&nbsp; '+s.label;
			$("dv_scores").appendChild(sc);
			sc.style.borderColor="#"+s.color;
			if(renderHalfWidth){
				sc.style.lineHeight="auto";
				sc.style.height=(2*sc.getHeight()-10)+"px";
				sc.style.width="40%";
				sc.style.cssFloat="left";
			} else {
				sc.style.lineHeight=(sc.getHeight()-4)+"px";
				sc.style.height=(sc.getHeight()-4)+"px";
			}
			if(canUpdate){
				sc.style.cursor="pointer";
				sc.onclick=function(){ DropViewer.setScore(sc); }
			}
		});
		if(canUpdate){
			let lbl=document.createElement("label");
			let cb=document.createElement("input");
			cb.id="dv_advanceonscore";
			cb.type="checkbox";
			lbl.appendChild(cb);
			lbl.innerHTML+=" Go to next drop after scoring";
			lbl.style.borderRadius=0;
			lbl.style.clear="both";
			$("dv_scores").appendChild(lbl);
			lbl.style.lineHeight=lbl.previous().getHeight()+"px";
			lbl.style.textAlign="center";
			lbl.style.fontWeight="bold";
			window.setTimeout(function(){
				if(parseInt(UserConfig.get("dropviewer_advanceOnScore","1"))===1){
					$("dv_advanceonscore").checked="checked";
				}
				$("dv_advanceonscore").observe("click",DropViewer.setAdvanceOnScore);
			},50);
		}
	},
	getScoringSystem_onFailure:function(transport){
		AjaxUtils.checkResponse(transport);
	},

	setAdvanceOnScore:function(){
		if($("dv_advanceonscore").checked){
			UserConfig.set("dropviewer_advanceOnScore",1);
		} else {
			UserConfig.set("dropviewer_advanceOnScore",0);
		}
	},

	setScoreByHotkey:function(hotkey){
		if(DropInfo==currentPanel){
			return DropInfo.setScoreByHotkey(hotkey);
		}
		$$("#dv_scores .dv_score").each(function(s){
			if(s.dataset.hotkey==hotkey){
				DropViewer.setScore(s);
			}
		});
	},

	setScore:function(scoreElem){
		if(scoreElem.hasClassName("updating")){ return false; }
		if(!canUpdate){ 
			alert("You don't have permission to update this plate. Talk to your administrator.");
			return false; 
		}

		let imageToScore=DropViewer.images[DropViewer.currentIndex];
		let imageId=imageToScore.id;
		let scoreId;

		if(scoreElem.hasClassName("dv_currentscore")){
			//unscore the drop altogether
			if(!confirm("Remove the score for this image?")){
				return false;
			}			
			scoreId=0;
			scoreElem.removeClassName("dv_currentscore");
			imageToScore.latestcrystalscoreid="";
		} else {
			//set the new score
			scoreElem.style.backgroundColor=scoreElem.style.borderColorLeft;
			scoreElem.addClassName("dv_currentscore");
			imageToScore.latestcrystalscoreid=scoreElem.dataset.scoreid;
			scoreId=scoreElem.dataset.scoreid;
		}
		//highlight it, unhighlight the rest
		scoreElem.addClassName("updating");
		new Ajax.Request('/api/dropimage/'+imageId, {
			method:'patch',
			parameters:{
				csrfToken:csrfToken,
				latestcrystalscoreid:scoreId
			},
			onSuccess:DropViewer.setScore_onSuccess,
			onFailure:DropViewer.setScore_onFailure,
		});

		let aos=$("dv_advanceonscore");
		if(aos && aos.checked){
			window.setTimeout(function(){ DropViewer.goToNext(true)},200);
		}
	},
	setScore_onSuccess:function(transport){
		AjaxUtils.checkResponse(transport);
		DropViewer.highlightCurrentImageScore();
	},
	setScore_onFailure:function(transport){
		if(transport.responseJSON && transport.responseJSON.error  && transport.responseJSON.error.indexOf("Duplicate entry") >-1){ 
			//Scored the same drop twice with same score in same second - just eat it. 
			return false; 
		}
		AjaxUtils.checkResponse(transport);
		DropViewer.highlightCurrentImageScore();
	},
	
	increaseMovieDelay:function(){
		DropViewer.changeMovieDelay(1);
	},
	decreaseMovieDelay:function(){
		DropViewer.changeMovieDelay(-1);
	},
	changeMovieDelay:function(direction){
		let md=DropViewer.movieDelay;
		md=(parseFloat(md)+(0.1*direction)).toFixed(1);
		if(md<0.1){ md=0.1; }
		UserConfig.set("dropviewer_movieDelay",md);
		$("dv_delay").innerHTML=md;
		$("tc_delay").innerHTML=md;
		DropViewer.movieDelay=md;
		if(DropViewer.movieInterval){
			DropViewer.stopMovie();
			DropViewer.startMovie();
		}
	},
	
	switchTo:function(){
		setCurrentModeButton($("viewbutton_dv"));
		if(DropViewer==currentPanel){ return false; }
		currentPanel=DropViewer;
		DropViewer.setCurrentImage(DropViewer.currentIndex);
		$$(".controlpanel").each(function(p){
			p.style.display="none";
		});
		$("dropviewerpanel").style.display="block";
		DropViewer.highlightCurrentImageScore();
		return true;
	},

	highlightCurrentImageScore:function(){
		let currentImageScore=DropViewer.images[DropViewer.currentIndex]['latestcrystalscoreid'];
		let scores=$$("#dv_scores .dv_score");
		if(!scores.length){
			window.setTimeout(DropViewer.highlightCurrentImageScore,100);
			return false;
		}
		scores.each(function(s){
			s.removeClassName("updating");
			if(s.dataset.scoreid==currentImageScore){
				s.style.backgroundColor=s.style.borderColor;
				s.addClassName("dv_currentscore");
			} else {
				s.style.backgroundColor="black";
				s.removeClassName("dv_currentscore");
			}
		});
	},
	
	switchFrom:function(){
		DropViewer.stopMovie();
	},
	
	goToFirst:function(){
		DropViewer.stopMovie();
		DropViewer.setCurrentImage(0);
	},
	goToPrevious:function(byDrop){
		DropViewer.stopMovie();
		if(byDrop || KeyboardShortcuts.ctrlPressed){
			DropViewer.previousImage();
		} else if(KeyboardShortcuts.shiftPressed){
			DropViewer.previousImageAboveScoreThreshold();
		} else {
			DropViewer.goLeftOneColumn();
		}
	},
	goToNext:function(byDrop){
		DropViewer.stopMovie();
		if(byDrop || KeyboardShortcuts.ctrlPressed){
			DropViewer.nextImage();
		} else if(KeyboardShortcuts.shiftPressed){
			DropViewer.nextImageAboveScoreThreshold();
		} else {
			DropViewer.goRightOneColumn();
		}
	},
	goToLast:function(){
		DropViewer.stopMovie();
		DropViewer.setCurrentImage(DropViewer.images.length-1);
	},

	goLeftOneColumn:function(){
		let current=DropViewer.getCurrentImage();
		let newIndex=DropViewer.getImageIndexByRowColumnDrop(1*current.row,1*current.col-1,1*current.dropnumber);
		if(null!==newIndex){ 
			DropViewer.setCurrentImage(newIndex);
			return;
		}
		for(let c=DropViewer.plateType.cols;c>=1;c--){
			newIndex=DropViewer.getImageIndexByRowColumnDrop(1*current.row-1,c,1*current.dropnumber);
			if(null!==newIndex){ 
				DropViewer.setCurrentImage(newIndex); 
				return;
			}
		}
		
	},
	goRightOneColumn:function(){
		let current=DropViewer.getCurrentImage();
		let newIndex=DropViewer.getImageIndexByRowColumnDrop(1*current.row,1*current.col+1,1*current.dropnumber);
		if(null!==newIndex){ 
			DropViewer.setCurrentImage(newIndex); 
			return;
		}
		for(let c=1;c<=DropViewer.plateType.cols;c++){
			newIndex=DropViewer.getImageIndexByRowColumnDrop(1*current.row+1,c,1*current.dropnumber);
			if(null!==newIndex){ 
				DropViewer.setCurrentImage(newIndex); 
				return;
			}
		}
	},
	goUpOneRow:function(){
		let current=DropViewer.getCurrentImage();
		let newIndex=DropViewer.getImageIndexByRowColumnDrop(1*current.row-1,1*current.col,1*current.dropnumber);
		if(null!==newIndex){ DropViewer.setCurrentImage(newIndex); }
	},
	goDownOneRow:function(){
		let current=DropViewer.getCurrentImage();
		let newIndex=DropViewer.getImageIndexByRowColumnDrop(1*current.row+1,1*current.col,1*current.dropnumber);
		if(null!==newIndex){ DropViewer.setCurrentImage(newIndex); }
	},
	
	getImageIndexByRowColumnDrop:function(rowNum,colNum,dropNum){
		let numImages=DropViewer.images.length;
		for(let i=0;i<numImages;i++){
			let img=DropViewer.images[i];
			if(img.row==rowNum && img.col==colNum && img.dropnumber==dropNum){
				return i;
			}
		}
		return null;
	},
	
	playOrStop:function(){
		if(DropViewer.movieInterval){
			DropViewer.stopMovie();
		} else {
			DropViewer.startMovie();
		}
	},
	startMovie:function(){
		if(DropViewer.currentIndex>=DropViewer.maxIndex){
			DropViewer.setCurrentImage(0);
		}
		DropViewer.movieInterval=window.setInterval(DropViewer.nextImage,DropViewer.movieDelay*1000);
		let ps=$("dv_playstop");
		ps.title="Stop";
		ps.addClassName("playing");
		$("dv_timecourseends").style.display="none";
	},
	stopMovie:function(){
		window.clearInterval(DropViewer.movieInterval);
		DropViewer.movieInterval=null;
		let ps=$("dv_playstop");
		ps.title="Play";
		ps.removeClassName("playing");
		DropViewer.getCurrentImageTimeCourse();
	},
	nextImage:function(){
		DropViewer.currentIndex++;
		if(!DropViewer.setCurrentImage(DropViewer.currentIndex)){
			DropViewer.currentIndex--;
			DropViewer.stopMovie();
		} else {
			DropViewer.preCacheMovieImages();
		}
	},
	previousImage:function(){
		DropViewer.currentIndex--;
		if(!DropViewer.setCurrentImage(DropViewer.currentIndex)){
			DropViewer.currentIndex++;
			DropViewer.stopMovie();
		}
	},
	nextImageAboveScoreThreshold:function(){
		DropViewer.nextOrPreviousAboveScoreThreshold(1);
	},
	previousImageAboveScoreThreshold:function(){
		DropViewer.nextOrPreviousAboveScoreThreshold(-1);
	},
	nextOrPreviousAboveScoreThreshold:function(direction){
		let scoringSystemId=DropViewer.scores[0].crystalscoringsystemid;
		let defaultThreshold=1*(DropViewer.scores[DropViewer.scores.length-1].scoreIndex);
		let indexThreshold=1*(UserConfig.get("dropviewer_scorethreshold_ss"+scoringSystemId, defaultThreshold));
		let newIndex=-1;
		for(let i=DropViewer.currentIndex+direction;i<=DropViewer.maxIndex&&i>=0;i+=direction){
			let latestScoreId=DropViewer.images[i].latestcrystalscoreid;
			if(DropViewer.scoreIndicesByScoreId["s"+latestScoreId] >= indexThreshold){
				newIndex=i;
				break;
			}
		}
		DropViewer.setCurrentImage(newIndex);
	},

	preCacheMovieImages:function(){
		let imagesToCache=25;
		let start=DropViewer.currentIndex;
		let end=DropViewer.currentIndex+imagesToCache;
		for(let i=start;i<end;i++){
			let img=DropViewer.images[i];
			if(img && !$("imagecache").down("#cache"+img.id)){
				$("imagecache").innerHTML+='<img alt=""' +
					'id="cache'+img.id+'" src="'+img.fullimageurl+'" style="height:2px" />';
				window.setTimeout(function(){
					if(img && $("cache"+img.id)){
						$("cache"+img.id).remove();
					}
				}, 10000);
			}
		}
	},
	
	getCurrentImage:function(){
		return DropViewer.images[DropViewer.currentIndex];
	},
	
	setCurrentImage:function(index){
		index*=1;
		if(index>DropViewer.maxIndex){ return false; }
		if(index<0){ return false; }
		let img=DropViewer.images[index];
		$("currentimage").src=img.fullimageurl;
		$("currentimage").dataset.micronshigh=img.pixelheight*img.micronsperpixely+"";
		DropViewer.fitImageToPane(img);
		Measure.setScale();
		DropViewer.setDropNameInTitle(img);
		DropViewer.currentIndex=index;
		DropViewer.highlightCurrentImageScore();
		DropViewer.getCurrentImageTimeCourse();
		DropInfo.refresh();
		return true;
	},

	fitImageToPane:function(){
		let img=DropViewer.images[DropViewer.currentIndex];
		if(img.pixelwidth/$("imagepane").getWidth() > img.pixelheight/$("imagepane").getHeight()){
			$("currentimage").style.width="100%";
			$("currentimage").style.height="auto";
		} else {
			$("currentimage").style.width="auto";
			$("currentimage").style.height="100%";
		}
		Measure.setScale();
		if(CrystalSelection==currentPanel){
			CrystalSelection.redrawCrosshairs();
		}
	},
	
	getWellNameForImage:function(img){
		let col=img.col*1;
		if(col<10){ col="0"+col; }
		return DropViewer.rowLabels[img.row] + col;
	},
	
	setDropNameInTitle:function(img){
		let dropName=DropViewer.getWellNameForImage(img)+"."+img.dropnumber;
		$("dropname").innerHTML=dropName;
		document.location.hash=dropName;
	},
	
	/**
	 * Retrieves all the image records for the current drop, as JSON. 
	 * Success handler attaches them to the current drop and updates first/last image.
	 */
	getCurrentImageTimeCourse:function(){
		//use a short timeout to break the call chain, otherwise pre-nav image's timecourse is fetched
		window.setTimeout(DropViewer._doGetCurrentImageTimeCourse,25);
	},
	_doGetCurrentImageTimeCourse:function(){
		let index=DropViewer.currentIndex;
		let obj=DropViewer.images[index];
		new Ajax.Request('/api/welldrop/'+obj.welldropid+'/timecourseimage?pagesize=1000',{
			method:'get',
			onSuccess:function(transport){ DropViewer.getCurrentImageTimeCourse_onSuccess(transport, DropViewer.currentIndex); },
			onFailure:function(transport){ DropViewer.getCurrentImageTimeCourse_onFailure(); },
		});
	},
	getCurrentImageTimeCourse_onSuccess:function(transport, imageIndex){
		let tc=transport.responseJSON.rows;
		DropViewer.images[imageIndex].timeCourseImages=tc;
		let first=$("dv_timecoursefirst");
		let last=$("dv_timecourselast");
		first.src=tc[0].thumbnailurl;
		first.dataset.fullsrc=tc[0].fullimageurl;
		first.onmouseover=DropViewer.swapInTimeCourseFirstImage;
		first.onmouseout=DropViewer.swapOutTimeCourseFirstImage;
		last.src=tc[tc.length-1].thumbnailurl;
		last.dataset.fullsrc=tc[tc.length-1].fullimageurl;
		last.onmouseover=DropViewer.swapInTimeCourseLastImage;
		last.onmouseout=DropViewer.swapOutTimeCourseLastImage;
		$("dv_timecourseends").style.display="block";
	},
	getCurrentImageTimeCourse_onFailure:function(){
	},

	swapInTimeCourseFirstImage: function(){ DropViewer.swapInTimeCourseEndImage($("dv_timecoursefirst"));  },
	swapOutTimeCourseFirstImage:function(){ DropViewer.swapOutTimeCourseEndImage($("dv_timecoursefirst")); },
	swapInTimeCourseLastImage:  function(){ DropViewer.swapInTimeCourseEndImage($("dv_timecourselast"));   },
	swapOutTimeCourseLastImage: function(){ DropViewer.swapOutTimeCourseEndImage($("dv_timecourselast"));  },
	
	swapInTimeCourseEndImage:function(img){
		if(DropViewer!=currentPanel){ return false; }
		if(""!=img.dataset.oldsrc && undefined!=img.dataset.oldsrc){ return false; /* because this is a repeat onkeydown while the key is held */}
		img.dataset.oldsrc=$("currentimage").src;
		$("currentimage").src=img.dataset.fullsrc;
	},
	swapOutTimeCourseEndImage:function(img){
		if(DropViewer!=currentPanel){ return false; }
		$("currentimage").src=img.dataset.oldsrc;
		img.dataset.oldsrc='';
	},
	
	olderImage:function(){
		DropViewer.olderOrNewerImage(-1);
	},
	newerImage:function(){
		DropViewer.olderOrNewerImage(1);
	},
	olderOrNewerImage:function(direction){
		if(DropViewer!=currentPanel){ return false; }
		let img=DropViewer.images[DropViewer.currentIndex];
		if(!img.timeCourseImages){ return false; }
		let timeCourseIndex=-1;
		let currentSrc=$("currentimage").src;
		for(let i=0; i<img.timeCourseImages.length; i++){
			if(currentSrc.indexOf(img.timeCourseImages[i].fullimageurl)!=-1){
				timeCourseIndex=i;
				break;
			}
		}
		timeCourseIndex+=direction;
		if(timeCourseIndex>=0 && timeCourseIndex<img.timeCourseImages.length){
			$("currentimage").src=img.timeCourseImages[timeCourseIndex].fullimageurl;
		}
	}
	
	
};




/**********************************************************************************************************
 * TIME COURSE 
 **********************************************************************************************************/

window.TimeCourse={

		//holds the src of the image showing before the time course was opened
		imageBefore:null,

		images:null,
		
		activeLightTypes:[],
		
		switchTo:function(){
			setCurrentModeButton($("viewbutton_tc"));
			TimeCourse.imageBefore=$("currentimage").src.slice(0); //copy it
			if(TimeCourse==currentPanel){ return false; }
			$$(".controlpanel").each(function(p){
				p.style.display="none";
			});
			$("timecoursepanel").style.display="block";
			currentPanel=TimeCourse;
			let container=$("tc_images");
			container.innerHTML="Waiting for images...";
			TimeCourse.writeTimeCourseImages();
			return true;
		},
		switchFrom:function(){
			$("currentimage").src=TimeCourse.imageBefore;
			Measure.setScale();
		},

		writeTimeCourseImages:function(){
			let currentImage=DropViewer.getCurrentImage();
			while(!currentImage.timeCourseImages){
				//we wait for the DOM to catch up
			}
			let container=$("tc_images");
			container.innerHTML="";
			TimeCourse.images=currentImage.timeCourseImages;
			if(TimeCourse.images.length===0){ return; }

			let regex=/[-:\s]/g;
			let parts=TimeCourse.images[0].imageddatetime.split(regex);
			let timeZero= new Date(parts[0], parts[1]-1, parts[2], parts[3], parts[4], 0).getTime();
			timeZero=Math.floor(timeZero/1000);

			let scores=DropViewer.scores;
			let scoreIdsToScores=[];
			
			currentImage.timeCourseImages.each(function(img){
				let light=img.lighttype;
				let d=document.createElement("div");
				d.onmouseover=function(){ TimeCourse.showTimePointImage(d); };
				d.onmouseout=function(){ TimeCourse.unshowTimePointImage(d); };
				d.onclick=function(){ TimeCourse.setCurrentImage(d, true); };
				d.image=img;
				d.className="timepoint tc_"+light;
				if(img.id==DropViewer.images[DropViewer.currentIndex].id){
					d.className+=" currenttimepoint";
				}
				let i=document.createElement("img");
				i.src=img.thumbnailurl;

				if(""!=img.latestcrystalscoreid){
					let scoreId=img.latestcrystalscoreid;
					if(!scoreIdsToScores["s"+scoreId]){
						scores.each(function(s){
							scoreIdsToScores["s"+s.id]=s;
						});
					}
					i.style.borderColor="#"+scoreIdsToScores["s"+scoreId]["color"];
					i.title="Score: "+scoreIdsToScores["s"+scoreId]["label"];
				}
				d.appendChild(i);

				parts=img.imageddatetime.split(regex);
				let offsetTime= new Date(parts[0], parts[1]-1, parts[2], parts[3], parts[4], 0).getTime();
				offsetTime=Math.floor(offsetTime/1000);
				let s=document.createElement("span");
				s.className="tc_dateoffset";
				s.innerText=ui.secondsToFriendlyUnits(offsetTime-timeZero);
				d.appendChild(s);
				let s2=document.createElement("span");
				s2.className="tc_actualdate";
				s2.innerText=ui.friendlyDate(img.imageddatetime);
				d.appendChild(s2);
				container.appendChild(d);
 				if(!$("tc_"+light)){
					let btn=document.createElement("div");
					btn.className="tc_lighttype tc_lighttypeactive";
					btn.id="tc_"+light;
					btn.innerText=light;
					btn.dataset.light=light;
					btn.onclick=function(){ TimeCourse.toggleLightTypeImages(btn) };
					btn.style.backgroundImage='url(/images/icons/lighttypes/'+light+'.png)';
					$("tc_lighttypes").appendChild(btn);
					TimeCourse.activeLightTypes.push(light);
				}
			});
		},

		toggleLightTypeImages:function(button){
			if($(button).hasClassName("tc_lighttypeactive")){
				$(button).removeClassName("tc_lighttypeactive");
				$$("#tc_images .tc_"+button.dataset.light).each(function(dv){
					//dv.style.display="none";
					dv.addClassName("suppressed");
				});
				TimeCourse.activeLightTypes=TimeCourse.activeLightTypes.without(button.dataset.light);
			} else {
				$(button).addClassName("tc_lighttypeactive");
				$$("#tc_images .tc_"+button.dataset.light).each(function(dv){
					//dv.style.display="block";
					dv.removeClassName("suppressed");
				});
				TimeCourse.activeLightTypes.push(button.dataset.light);
			}
			let curr=$("tc_images").down(".currenttimepoint");
			if(curr && curr.hasClassName("suppressed")){
				if(!TimeCourse.goToPrevious()){
					TimeCourse.goToNext();
				}
			}
		},

		/**
		 * Mouseover/out image swap
		 */
		showTimePointImage:function(elem){
			if(!elem.image){ return false; }
			elem.previousImage=$("currentimage").src;
			$("currentimage").src=elem.image.fullimageurl;
		},		
		unshowTimePointImage:function(elem){
			if(elem.hasClassName("currenttimepoint")){ return false; }
			if(!elem.previousImage){ return false; }
			$("currentimage").src=elem.previousImage;
		},
		
		/**
		 * Movie/navigation controls
		 */
		goToFirst:function(){
			TimeCourse.stopMovie();
			let first=$("tc_images").down("div.timepoint:not(.suppressed)");
			if(!first){ return false; }
			TimeCourse.setCurrentImage(first);
		},
		goToPrevious:function(){
			TimeCourse.stopMovie();
			return TimeCourse.previousImage();
		},
		goToNext:function(){
			TimeCourse.stopMovie();
			return TimeCourse.nextImage();
		},
		goToLast:function(){
			let active=$$("#tc_images div.timepoint:not(.suppressed)");
			if(!active || 0===active.length){ return false; }
			let last=active.last();
			TimeCourse.setCurrentImage(last);
		},
		playOrStop:function(){
			if(TimeCourse.movieInterval){
				TimeCourse.stopMovie();
			} else {
				TimeCourse.startMovie();
			}
		},
		startMovie:function(){
			let playStop=$("tc_playstop");
			let current=$("tc_images").down(".currenttimepoint");
			let next=current.next("div.timepoint:not(.suppressed)");
			if(!next){
				TimeCourse.goToFirst();
			}
			TimeCourse.movieInterval=window.setInterval(TimeCourse.nextImage,DropViewer.movieDelay*1000);
			playStop.title="Stop";
			playStop.addClassName("playing");
		},
		stopMovie:function(){
			window.clearInterval(TimeCourse.movieInterval);
			TimeCourse.movieInterval=null;
			let playStop=$("tc_playstop");
			playStop.title="Play";
			playStop.removeClassName("playing");
		},
		nextImage:function(){
			let current=$("tc_images").down(".currenttimepoint");
			let next=current.next("div.timepoint:not(.suppressed)");
			if(!next){
				TimeCourse.stopMovie();
				return false;
			}
			TimeCourse.setCurrentImage(next);
			return true;
		},
		previousImage:function(){
			let current=$("tc_images").down(".currenttimepoint");
			let prev=current.previous("div.timepoint:not(.suppressed)");
			if(!prev){
				TimeCourse.stopMovie();
				return false;
			}
			TimeCourse.setCurrentImage(prev);
			return true;
		},
		olderImage:function(){
			TimeCourse.previousImage();
		},
		newerImage:function(){
			TimeCourse.nextImage();
		},
		
		getCurrentImage:function(){
			return TimeCourse.images[TimeCourse.currentIndex];
		},
		
		setCurrentImage:function(container, dontScroll){
			let tcImages=$("tc_images");
			let currentImage=$("currentimage");
			if(tcImages.down(".currenttimepoint")){
				tcImages.down(".currenttimepoint").removeClassName("currenttimepoint");
			}
			container.addClassName("currenttimepoint");
			currentImage.src=container.image.fullimageurl;
			if(!dontScroll){
				tcImages.scrollTop=container.offsetTop;
			}
			currentImage.dataset.micronshigh=container.image.pixelheight*container.image.micronsperpixely+"";
			Measure.setScale();
			return true;
		},

};



/**********************************************************************************************************
 * DROP INFORMATION PANEL
 **********************************************************************************************************/

window.DropInfo={

	/**
	 * Which items are shown on the info panel.
	 * This is the default state, which will be overwritten by user's configuration
	 * (or become the user's configuration, if none exists).
	 */
	infoPanelItems:{
			'dropviewer_infopanel_show_Drop_scoring':0,
			'dropviewer_infopanel_show_Protein':1,
			'dropviewer_infopanel_show_Well_solution':1,
			'dropviewer_infopanel_show_Protein_solution':1,
			'dropviewer_infopanel_show_Drop_volumes':1,
			'dropviewer_infopanel_show_Plate_details':1,
			'dropviewer_infopanel_show_Incubation':1,
	},
	
	switchTo:function(){
		Object.keys(DropInfo.infoPanelItems).each(function(k){
			DropInfo.infoPanelItems[k]=parseInt(UserConfig.get(k, DropInfo.infoPanelItems[k]));
		});
		
		setCurrentModeButton($("viewbutton_di"));
		$$(".controlpanel").each(function(p){
			p.style.display="none";
		});
		currentPanel=DropInfo;
		DropInfo.render();
		return true;
	},
	switchFrom:function(){
		$("dropinfopanel").innerHTML='';
	},
	refresh:function(){
		if(!DropViewer.images[DropViewer.currentIndex].welldrop){
			DropInfo.getWellDropInfo(); //triggers render()
		} else {
			DropInfo.render();
		}
	},
	getWellDropInfo:function(){
		let wellDropId=DropViewer.images[DropViewer.currentIndex].welldropid;
		new Ajax.Request("/api/welldrop/"+wellDropId,{
			method:"get",
			onSuccess:DropInfo.getWellDropInfo_onSuccess,
			onFailure:DropInfo.getWellDropInfo_onFailure,
		});
	},
	getWellDropInfo_onSuccess:function(transport){
		//set response to current drop's "welldrop" property
		if(!AjaxUtils.checkResponse(transport)){ return false; }
		DropViewer.images[DropViewer.currentIndex].welldrop=transport.responseJSON;
		DropInfo.render();
	},
	getWellDropInfo_onFailure:function(){
		DropInfo.render();
	},
	render:function(){
		if(DropInfo!==currentPanel){ return false; }
		if(!DropViewer.scores){
			window.setTimeout(DropInfo.render, 250);
			return false;
		}
		let wellDrop=DropViewer.images[DropViewer.currentIndex].welldrop;
		let img=DropViewer.images[DropViewer.currentIndex];
		let screen=DropViewer.screen;
		let out='';

		if(1==DropInfo.infoPanelItems.dropviewer_infopanel_show_Drop_scoring){
			let currentScoreColor='';
			out+='<h2>Drop score</h2><div style="border-left:1.5em solid transparent"><select id="dropinfo_scoring">';
			if(canUpdate){
				out+='<option value="0" onclick="DropInfo.setScoreByWidget()">(unscored)</option>';
				DropViewer.scores.each(function(sc){
					let selected='';
					if(sc.id==img.latestcrystalscoreid){ 
						selected='selected="selected"'; 
						currentScoreColor=sc.color;
					}
					out+='<option '+selected+' value="'+sc.id+'" onclick="DropInfo.setScoreByWidget()" style="padding-left:0.25em;border-left:0.5em solid #'+sc.color+'">['+sc.hotkey+'] '+sc.label+'</option>';
				});
				out+='</select></div>';
				//highlight current
			} else {
				out+='<div style="padding-left:0.25em;border-left:0.5em solid #'+sc.color+'">'+sc.label+'</div>'
			}
		}
		
		if(1==DropInfo.infoPanelItems.dropviewer_infopanel_show_Protein){
			out+='<br/><h2>Protein</h2>';
			if(wellDrop && wellDrop.proteinname && wellDrop.proteinname!==""){
				out+=DropViewer.images[DropViewer.currentIndex].welldrop.proteinname;
				out+='<br/>Construct: '+DropViewer.images[DropViewer.currentIndex].welldrop.constructname;
			} else {
				out+=DropViewer.warningImage+'Not set';
			}
		}

		if(1==DropInfo.infoPanelItems.dropviewer_infopanel_show_Well_solution){
			out+='<br/><br/><h2>Well solution</h2>';
			if(null==screen || !screen.conditions){
				out+='Screen: (none chosen)';
			} else {
				let screenDescription='(Not found)';
				let screenRow=1*img.row+1*data.offsetinscreeny;
				let screenCol=1*img.col+1*data.offsetinscreenx;
				screen.conditions.each(function(c){
					if(1*c.col==screenCol && 1*c.row==screenRow){
						screenDescription=c.description;
					}
				});
				out+=screenDescription;
				out+='<br/>Screen: <a href="/screen/'+screen.id+'">'+screen.name+'</a>';
			}
		}
		
		if(1==DropInfo.infoPanelItems.dropviewer_infopanel_show_Protein_solution){
			out+='<br/><br/><h2>Protein solution</h2>';
			if(""===wellDrop.proteinconcentrationamount || ""===wellDrop.proteinconcentrationunit){
				out+='Protein concentration: '+DropViewer.warningImage+'Not set';

			} else {
				let unit=wellDrop.proteinconcentrationunit;
				unit=unit.replace("uL","&#181;L");
				unit=unit.replace("ug","&#181;g");
				out+='Protein concentration: '+wellDrop.proteinconcentrationamount + unit;
			}
			out+='<br/>';
			if(""===wellDrop.proteinbuffer){
				out+='Protein buffer: '+DropViewer.warningImage+'Not set';
			} else {
				out+='Protein buffer: '+wellDrop.proteinbuffer;
			}
		}
		
		if(1==DropInfo.infoPanelItems.dropviewer_infopanel_show_Drop_volumes){
			out+='<br/><br/><h2>Drop volumes</h2>';
			if(""===wellDrop.proteinsolutionamount || ""===wellDrop.proteinsolutionunit){
				out+='Protein solution: '+DropViewer.warningImage+'Not set';
			} else {
				let unit=wellDrop.proteinsolutionunit;
				unit=unit.replace("uL","&#181;L");
				unit=unit.replace("ug","&#181;g");
				out+='Protein solution: '+wellDrop.proteinsolutionamount + unit;
			}
			out+='<br/>';
			if(""===wellDrop.wellsolutionamount || ""===wellDrop.wellsolutionunit){
				out+='Well solution: '+DropViewer.warningImage+'Not set';
			} else {
				let unit=wellDrop.wellsolutionunit;
				unit=unit.replace("uL","&#181;L");
				unit=unit.replace("ug","&#181;g");
				out+='Well solution: '+wellDrop.wellsolutionamount + unit;
			}
		}
			
		if(1==DropInfo.infoPanelItems.dropviewer_infopanel_show_Plate_details){
			out+='<br/><br/><h2>Plate</h2>';
			out+='Type: <a href="/platetype/'+DropViewer.plateType.id+'">'+DropViewer.plateType.name+'</a>';
			out+='<br/>Barcode: <a href="/plate/'+data.plateid+'">'+data.platename+'</a>';
			out+='<br/>Description: '+data.platedescription;
		}
		
		if(1==DropInfo.infoPanelItems.dropviewer_infopanel_show_Incubation){
			out+='<br/><br/><h2>Incubation</h2>';
			out+='Imager: <a href="/imager/'+data.imagerid+'">'+data.imagerfriendlyname+'</a>';
			out+='<br/>Temperature: '+data.temperature+'&deg;C';
		}
		
		$("dropinfopanel").innerHTML=out;
		$("dropinfopanel").style.display="block";
		if($("dropinfo_scoring") && ""!=currentScoreColor){
			$("dropinfo_scoring").up("div").style.borderColor="#"+currentScoreColor;
		}
	},

	setScoreByHotkey:function(hotkey){
		if(DropInfo!==currentPanel){
			return false;
		}
		let scoreElem=$("dropinfo_scoring");
		DropViewer.scores.each(function(s){
			if(s.hotkey==hotkey){
				if(scoreElem){
					scoreElem.value=s.id;
				}
				DropInfo.setScoreById(s.id);
			}
		});
	},
	setScoreByWidget:function(){
		let scoreElem=$("dropinfo_scoring");
		if(scoreElem.hasClassName("updating")){ return false; }
		DropInfo.setScoreById(scoreElem.value);	
	},
	
	setScoreById:function(scoreId){
		if(!canUpdate){ 
			alert("You don't have permission to update this plate. Talk to your administrator.");
			return false; 
		}

		let imageToScore=DropViewer.images[DropViewer.currentIndex];
		let imageId=imageToScore.id;
		let scoreElem=$("dropinfo_scoring");
		if(scoreElem){
			scoreElem.addClassName("updating");
			scoreElem.value=scoreId;
			let col='transparent';
			DropViewer.scores.each(function(sc){
				if(sc.id==scoreId){
					col="#"+sc.color;
				}
			});
			scoreElem.up("div").style.borderColor=col;
		}
		new Ajax.Request('/api/dropimage/'+imageId, {
			method:'patch',
			parameters:{
				csrfToken:csrfToken,
				latestcrystalscoreid:scoreId
			},
			onSuccess:DropInfo.setScore_onSuccess,
			onFailure:DropInfo.setScore_onFailure,
		});
		
	},
	setScore_onSuccess:function(transport){
		AjaxUtils.checkResponse(transport);
		let scoreElem=$("dropinfo_scoring");
		if(scoreElem){
			scoreElem.removeClassName("updating");
			scoreElem.up("div").removeClassName("updating");
		}
	},
	setScore_onFailure:function(transport){
		if(transport.responseJSON && transport.responseJSON.error  && transport.responseJSON.error.indexOf("Duplicate entry") >-1){ 
			//Scored the same drop twice with same score in same second - just eat it. 
			return false; 
		}
		AjaxUtils.checkResponse(transport);
		let scoreElem=$("dropinfo_scoring");
		if(scoreElem){
			scoreElem.removeClassName("updating");
			scoreElem.up("div").removeClassName("updating");
		}
	},

	
};

/**********************************************************************************************************
 * CRYSTAL SELECTION PANEL
 **********************************************************************************************************/

window.CrystalSelection={
		
	crystals:[],
	currentCrystalNumber:null,

	switchTo:function(){
		setCurrentModeButton($("viewbutton_xs"));
		Measure.stop();
		$$(".controlpanel").each(function(p){
			p.style.display="none";
		});
		$("selectpanel").style.display="block";
		currentPanel=CrystalSelection;
		CrystalSelection.getCrystals();
		CrystalSelection.getPinCategoryAndTypes();
		$("currentimage").observe("click", CrystalSelection.addCrystal);
		return true;
	},
	switchFrom:function(){
		$$(".xs_crosshair").each(function(ch){ 
			if("xs_crosshair"===ch.id){
				ch.style.display="none";
			} else {
				ch.remove(); 
			} 
		});
		$("currentimage").stopObserving("click", CrystalSelection.addCrystal);
		if(!Measure.suppressed){
			Measure.init();
		}
		$("selectpanel").innerHTML='';
		CrystalSelection.crystals=null;
		CrystalSelection.currentCrystalNumber=null;
	},
	refresh:function(){
		if(CrystalSelection!==currentPanel){ return false; }
		let sp=$("selectpanel");
		let out='';
		let currentCrystal=null;
		if(1===CrystalSelection.crystals.length){
			CrystalSelection.currentCrystalNumber=CrystalSelection.crystals[0].numberindrop;
		}
		CrystalSelection.crystals.each(function(c){
			if(c.numberindrop==CrystalSelection.currentCrystalNumber){
				currentCrystal=c;
				$("selectpanel").currentCrystal=c;
			}
		});
		CrystalSelection.redrawCrosshairs();
		
		//No crystal selected
		if(null==CrystalSelection.currentCrystalNumber){
			out+='<h3>Crystals</h3>';
			out+='<p>Click any of the crosshairs on the image to select it, or click anywhere on the image to select a new crystal.</p>';
			sp.innerHTML=out;
			sp.style.display="block";
			return;
		}
	
		//Crystal selected, show its info
		let markedOnCurrentImage=!($("currentimage").src.indexOf(currentCrystal.dropimageid)<1);

		let canDeleteThisCrystal=(canDeleteCrystals && 0===parseInt(currentCrystal.isfished));
		
		out='<br/><h2>Crystal '+CrystalSelection.currentCrystalNumber+' overview - <a href="/crystal/'+currentCrystal.id+'">View full details</a></h2>';
		out+='<div id="selectpaneltabs">';
		
		out+='<h3 id="xtaldetails">Details</h3>';
		out+='<div id="xtaldetails_body" class="infopanel">';
			if(!currentCrystal.isDummyCoordinate && !markedOnCurrentImage){
				out+='<p>Crystal was marked on an earlier or later image of this drop.</p><p id="movetoimagectrl"></p>';
			}
			out+='<form action="#" id="xtaldetailsform"></form>';
			if(canDeleteCrystals){
				if(canDeleteThisCrystal){
					out+='<br/><hr/><br/><div><input type="button" id="xtaldelete" onclick="CrystalSelection.deleteCurrentCrystal()" value="Delete crystal" /></div>';
				} else {
					out+='<br/><hr/><p>This crystal has been fished.</p><p>It cannot be deleted.</p>';
				}
			}
	
		out+='</div>';

		out+='<h3 id="xtalfiles">Files</h3>';
		out+='<div id="xtalfiles_body" class="infopanel">';
			out+='<div id="crystalfiles">Waiting for files...</div>';
		out+='</div>';
		
		out+='<h3 id="xtalnotes">Notes</h3>';
		out+='<div id="xtalnotes_body" class="infopanel">';
			out+='<div id="crystalnotes">Waiting for notes...</div>';
		out+='</div>';

		if(canUpdate && 0===parseInt(currentCrystal.isfished)){
			out+='<h3 id="xtalfish">QuickFish</h3>';
			out+='<div id="xtalfish_body" class="infopanel">';
			out+='<p>If you\'re just fishing the crystal onto a pin for testing, you can do that here. You will need a barcoded pin.</p>';
			out+='<p>If you\'re not using barcoded pins, or if you want to store the pin in a puck, you\'ll need to use the <a href="/containercontent/create">Crystal Fishing</a> page.</p>';
			out+='</div>';
		}
		
		sp.innerHTML=out;
		Crystal.writeDetailsFormFields(currentCrystal, canUpdate, $("xtaldetailsform"));

		sp.style.display="block";
		
		if($("nameseparator") && (""===currentCrystal.suffix || !currentCrystal.suffix)){
			$("nameseparator").style.visibility="hidden";
		}

		if(canUpdate){
			if($("movetoimagectrl")){
				$("movetoimagectrl").innerHTML='<input type="button" style="cursor:pointer" onclick="CrystalSelection.updateCrystalImageNumber()" id="movetocurrentimage" value="Move crystal mark to this image" title="Crosshair was marked on a different image of this drop."/>';
			}
		}
		
		//get all the crystals, plot them
		if(null!=CrystalSelection.currentCrystalNumber){
			CrystalSelection.crystals.each(function(c){
				if(c.numberindrop==CrystalSelection.currentCrystalNumber){
					CrystalSelection.getCrystalNotes(c);
					CrystalSelection.getCrystalFiles(c);
				}
			});
		}
		sp.select("h3").each(function(h){
			h.observe("click",CrystalSelection.switchTab);
		});
		sp.select(".infopanel").each(function(p){
			p.style.position="absolute";
			p.style.bottom="6em";
		});
		if(canUpdate && 0==currentCrystal.isfished){
			let ff=ui.textField({ name:'quickfish_barcode', label:'Scan pin barcode', value:'', },$("xtalfish_body"));
			ff.down("input").stopObserving("keyup");
			ff.down("input").stopObserving("click");
			ff.down("input").observe("keyup",CrystalSelection.startQuickFish);
			$("xtalfish").observe("click", function(){ff.down("input").focus(); } )

		}
		window.setTimeout(function(){
			CrystalSelection.switchTab();
		},50);
	},

	
	switchTab:function(evt){
		let sp=$("selectpanel");

		let clicked=sp.down("h3");
		if(evt){
			clicked=evt.target;
		} else if(CrystalSelection.currentTab){
			clicked=CrystalSelection.currentTab;
		}
		if(!clicked){ return; }
		let id=clicked.id;
		CrystalSelection.currentTab=clicked;
		sp.select("h3").each(function(h){
			h.removeClassName("current");
		});
		sp.select(".infopanel").each(function(p){
			p.style.display="none";
		});
		$(id).addClassName("current");
		$(id+"_body").style.display="block";
	},
	
	redrawCrosshairs:function(){
		$$(".xs_crosshair").each(function(x){
			if("xs_crosshair"===x.id){
				x.style.display="none";
			} else {
				x.remove();
			}
		});
		CrystalSelection.crystals.each(function(c){
			CrystalSelection.writeCrosshair(c);
		});
	},
	
	writeCrosshair: function(crystal){
		let img=DropViewer.getCurrentImage();
		let newCrosshair=$("xs_crosshair").clone(true);
		let scale=$("currentimage").getHeight() / img.pixelheight;
		let offset=25;
		newCrosshair.id="crystal"+crystal.numberindrop;
		newCrosshair.dataset.crystalid=crystal.id;
		let pixelx=(crystal.pixelx * scale)-offset;
		let pixely=(crystal.pixely * scale)-offset;
		newCrosshair.style.top=pixely+"px";
		newCrosshair.style.left=pixelx+"px";
		newCrosshair.style.display="block";
		newCrosshair.down(".xs_number").innerHTML=crystal.numberindrop;
		$("xs_crosshair").up().appendChild(newCrosshair);
		newCrosshair.observe("click",CrystalSelection.showCrystal);
		newCrosshair.crystal=crystal;
		if(CrystalSelection.currentCrystalNumber==crystal.numberindrop){
			newCrosshair.addClassName("xs_currentcrosshair");
		} else if($("currentimage").src.indexOf(crystal.dropimageid)<1){
			newCrosshair.addClassName("xs_crosshairfromdifferentimage");
			newCrosshair.title="Crystal was marked on a different image of this drop";
		}
		return newCrosshair;
	},
	
	getPinCategoryAndTypes:function(){
		new Ajax.Request('/api/containercategory/name/Pin',{
			method:'get',
			onSuccess:function(transport){
				if(transport.responseJSON.rows){
					CrystalSelection.pinCategory=transport.responseJSON.rows[0];
					if(!CrystalSelection.pinCategory.userscancreate){ return; }
					new Ajax.Request('/api/containertype/containercategoryid/'+CrystalSelection.pinCategory.id, {
						method:'get',
						onSuccess:function(transport){
							if(transport.responseJSON.rows){
								CrystalSelection.pinTypes=transport.responseJSON.rows;
							}
						},
						onFailure:function(){}
					});
				}
			},
			onFailure:function(){}
		});
	},
	
	
	startQuickFish:function(evt){
		if(13!=evt.keyCode){ return; } //fire only on enter
		CrystalSelection.doStartQuickFish();
	},
	doStartQuickFish:function(){
		let barcodeField=$("quickfish_barcode");
		let barcode=barcodeField.value.trim();
		if(""===barcode){ return false; }
		barcodeField.up("label").addClassName("updating");
		new Ajax.Request("/api/container/name/"+barcode,{
			method:"get",
			onSuccess:function(transport){ 
				if(!transport.responseJSON || !transport.responseJSON.rows){
					return AjaxUtils.checkResponse(transport);
				}
				let pin=transport.responseJSON.rows[0];
				if("pin"!==pin.containercategoryname.toLowerCase()){
					CrystalSelection.doQuickFishError('Container with barcode '+barcode+' is a '+pin.containercategoryname.toLowerCase()+', not a pin.');
					return false;
				}
				CrystalSelection.quickFishGetPinContents(pin);
			},
			onFailure:function(transport){ 
				if(404===transport.status){
					//Pin does not exist
					if(!CrystalSelection.pinCategory || 1!=CrystalSelection.pinCategory.userscancreate || !CrystalSelection.pinTypes){
						//Can't create pin - insufficient info
						CrystalSelection.doQuickFishError('No pin found with barcode '+barcode+'.');
						return false; 
					} else if(1===CrystalSelection.pinTypes.length){
						//Only one type of pin - create it
						CrystalSelection.quickFishCreatePin(barcode, CrystalSelection.pinTypes[0].id);
						return false; 
					} else {
						//More than one type of pin - user must choose
						CrystalSelection.showPinTypeSelectionDialog(barcode);
						return false; 
					}
				}
				AjaxUtils.checkResponse(transport);
			},
		});
	},
	
	quickFishCreatePin:function(barcode, containerTypeId){
		ui.closeModalBox();
		new Ajax.Request('/api/container',{
			method:'post',
			parameters:{
				csrfToken:csrfToken,
				name:barcode,
				containertypeid:containerTypeId
			},
			onSuccess:CrystalSelection.doStartQuickFish,
			onFailure:function(){ alert("Pin does not exist, and it could not be created."); }
		});
	},

	showPinTypeSelectionDialog:function(barcode){
		ui.modalBox({
			title:"Icebear has no pin with barcode "+barcode+". What kind of pin is it?",
			content:'<div id="fish_ct_pin"></div></div>'
		});
		CrystalSelection.pinTypes.each(function(ct){
			let catName=ct.containercategoryname.toLowerCase();
			let categoryDiv=$("fish_ct_"+catName);
			if(!categoryDiv){ return; } //Not interested in this containercategory
			let html='';
			html+='<div class="fish_containertype" data-containertypeid="'+ct.id+'" data-barcode="'+barcode+
				'" onclick="CrystalSelection.quickFishCreatePin(this.dataset.barcode, this.dataset.containertypeid)" style="cursor:pointer">';
			html+='<div class="fish_containertypename fish_'+catName+'">'+ct.name+'</div>';
			html+='</div>';
			categoryDiv.innerHTML+=html;
		});
		
	},
	
	doQuickFishError:function(err){
		alert(err);
		let bc=$("quickfish_barcode");
		bc.up("label").removeClassName("updating");
		bc.value="";
		bc.focus();
	},
	
	quickFishGetPinContents:function(pin){
		new Ajax.Request("/api/container/"+pin.id+'/content',{
			method:"get",
			onSuccess:function(){
				//Success is failure! The pin is full.
				CrystalSelection.doQuickFishError('Pin '+pin.name+' has a crystal in it. Scan another pin.');
			},
			onFailure:function(transport){ 
				if(404===transport.status){
					//Failure is success! The pin is empty.
					let currentCrystal=null;
					CrystalSelection.crystals.each(function(c){
						if(c.numberindrop==CrystalSelection.currentCrystalNumber){
							currentCrystal=c;
							$("selectpanel").currentCrystal=c;
						}
					});
					CrystalSelection.quickFishDoFish(pin, currentCrystal);
				} else {
					//Real, actual failure.
					CrystalSelection.doQuickFishError('Could not determine whether pin is full. Try another.');
				}
			}
		});
	},
	quickFishDoFish:function(pin,crystal){
		if(!pin||!crystal){
			CrystalSelection.doQuickFishError('Could not record crystal fishing.');
		}
		new Ajax.Request('/api/containercontent/',{
			method:'post',
			parameters:{
				'csrfToken':csrfToken,
				'parent':pin.id,
				'child':crystal.id,
				'position':1
			},
			onSuccess:function(){
				crystal.isfished=1;
				new Ajax.Request('/api/crystal/'+crystal.id,{
					method:'patch',
					parameters:{
						'csrfToken':csrfToken,
						'isfished':1
					},
					//assume worked, no further handling needed
				});
				let noteText="Crystal fished onto pin "+pin.name+".";
				let bc=$("quickfish_barcode");
				bc.up("label").removeClassName("updating");
				bc.style.display="none";
				let lbl=bc.up("label").down("span.label");
				lbl.style.cssFloat="none";
				lbl.innerHTML=noteText;
				window.setTimeout(function(){
					CrystalSelection.quickFishSaveNote(crystal.id, noteText);
				},1000);
			},
			onFailure:function(transport){
				AjaxUtils.checkResponse(transport);
			}
		});
	},
	
	quickFishSaveNote:function(crystalId, noteText){
		new Ajax.Request("/api/note",{
			method:'post',
			parameters:{
				'csrfToken':csrfToken,
				'parentid':crystalId,
				'text':noteText
			},
			onSuccess:function(){
				CrystalSelection.quickFishAfterSaveNote(crystalId);
			},
			onFailure:function(){
				alert("Crystal was fished, but note describing this could not be saved.");
				CrystalSelection.quickFishAfterSaveNote(crystalId);
			}
		});
	},
	
	quickFishAfterSaveNote:function(crystalId){
		CrystalSelection.crystals.each(function(c){
			if(c.id==crystalId){
				c.isfished=true;
			}
		});
		CrystalSelection.currentTab=$("selectpaneltabs").down("h3");
		CrystalSelection.refresh();
	},
	
	
	updateCrystalImageNumber:function(){
		let crystalId=0;
		let imageId=DropViewer.images[DropViewer.currentIndex].id;
		CrystalSelection.crystals.each(function(xtal){
			if(xtal.numberindrop==CrystalSelection.currentCrystalNumber){
				crystalId=xtal.id;
			}
		});
		if(0===crystalId){ alert("Could not find crystal ID"); return false; }
		$("movetocurrentimage").up("h2").addClassName("updating");
		new Ajax.Request('/api/crystal/'+crystalId,{
			method:'patch',
			parameters:{
				csrfToken:csrfToken,
				dropimageid:imageId
			},
			onSuccess:CrystalSelection.updateCrystalImageNumber_onSuccess,
			onFailure:CrystalSelection.updateCrystalImageNumber_onFailure,
		});
	},
	updateCrystalImageNumber_onSuccess:function(transport){
		let updated=transport.responseJSON.updated;
		CrystalSelection.crystals.each(function(xtal){
			if(xtal.id==updated.id){
				xtal.dropimageid=updated.dropimageid;
			}
		});
		let m=$("movetocurrentimage");
		m.up("h2").removeClassName("updating");
		m.remove();
	},
	updateCrystalImageNumber_onFailure:function(transport){
		$("movetocurrentimage").up("h2").removeClassName("updating");
		AjaxUtils.checkResponse(transport);
	},
	
	showCrystal:function(evt){
		let xh=evt.target;
		while(!xh.crystal){
			xh=xh.up();
		}
		CrystalSelection.currentCrystalNumber=xh.crystal.numberindrop;
		CrystalSelection.currentTab=null;
		CrystalSelection.refresh();
	},
	
	getCrystals:function(){
		let img=DropViewer.getCurrentImage();
		if(!img.crystals){
			new Ajax.Request('/api/welldrop/'+img.welldropid+'/crystal',{
				method:'get',
				onSuccess:CrystalSelection.getCrystals_onSuccess,
				onFailure:CrystalSelection.getCrystals_onFailure,
			});
		} else {
			CrystalSelection.crystals=img.crystals;
			CrystalSelection.refresh();
		}
	},
	getCrystals_onSuccess:function(transport){
		let img=DropViewer.getCurrentImage();
		img.crystals=transport.responseJSON.rows;
		CrystalSelection.crystals=transport.responseJSON.rows;
		let numCrystals=CrystalSelection.crystals.length;
		if(0===numCrystals){
			CrystalSelection.refresh();
			return;
		}
		let counter=0;
		CrystalSelection.crystals.each(function(c){
			new Ajax.Request('/api/crystal/'+c.id+'/diffractionrequest',{
				method:'get',
				onSuccess:function(transport){
					c.diffractionrequests=transport.responseJSON.rows;
					counter++;
					if(counter===numCrystals){ CrystalSelection.refresh(); }
				},
				onFailure:function(){
					counter++;
					if(counter===numCrystals){ CrystalSelection.refresh(); }
				}
			});
		});
		window.setTimeout(function(){
			if(!$("selectpanel").down("h3")){
				CrystalSelection.refresh();
			}
		},2000);
	},
	getCrystals_onFailure:function(transport){
		if(404===transport.status){
			CrystalSelection.crystals=[];
			CrystalSelection.refresh();
			return;
		}
		alert("Could not retrieve list of crystals for this well.");
	},


	getCrystalFiles:function(crystal){
		new Ajax.Request('/api/crystal/'+crystal.id+'/file',{
			method:'get',
			onSuccess:CrystalSelection.getCrystalFiles_onSuccess,
			onFailure:CrystalSelection.getCrystalFiles_onFailure,
		})
	},
	getCrystalFiles_onSuccess:function(transport){
		AjaxUtils.checkResponse(transport);
		let fb=$("xtalfiles_body");
		fb.innerHTML='';
		let out='';
		CrystalSelection.writeCrystalFilesForm();
		transport.responseJSON.rows.each(function(file){
			out+='<a href="/file/'+file.id+'" target="_blank">'+file.filename+'</a><br/>';
		});
		fb.innerHTML+=out;
	},
	getCrystalFiles_onFailure:function(transport){
		$("xtalfiles_body").innerHTML='';
		if(404===transport.status){
			CrystalSelection.writeCrystalFilesForm();
			return;
		}
		alert("Could not retrieve files for this crystal.");
	},
	writeCrystalFilesForm:function(){
		let out='';
		if(!canUpdate){ return; }
		out+='<form id="crystalfilesadd" method="post" action="/api/file" onsubmit="alert(\'!!\');return false">';
		out+='<input type="hidden" value="'+csrfToken+'" name="csrfToken" />';
		out+='<input type="hidden" value="'+$("selectpanel").currentCrystal.id+'" name="parentid" />';
		out+='<strong>Attach a file: </strong><input id="crystalfilesaddfile" type="file" name="file" onchange="ui.submitForm(this)" />';
		out+='</form>';
		out+='<hr/>';
		$("xtalfiles_body").innerHTML+=out;
		window.setTimeout(function(){ 
			$("crystalfilesaddfile").options={
				afterSuccess:function(){
					CrystalSelection.refresh();
				}
			}
		},50);
	},

	
	getCrystalNotes:function(crystal){
		new Ajax.Request('/api/crystal/'+crystal.id+'/note',{
			method:'get',
			onSuccess:CrystalSelection.getCrystalNotes_onSuccess,
			onFailure:CrystalSelection.getCrystalNotes_onFailure,
		})
	},
	getCrystalNotes_onSuccess:function(transport){
		AjaxUtils.checkResponse(transport);
		let nb=$("xtalnotes_body");
		nb.innerHTML='';
		let out='';
		CrystalSelection.writeCrystalNotesForm();
		transport.responseJSON.rows.each(function(note){
			out+='<h4><a href="/user/'+note.userid+'">'+note.username+'</a>, '+ui.friendlyDate(note.createtime)+'</h4>';
			out+=ui.urlifyAndBreakNoteText(note.text)+'<hr/><br/>';
		});
		nb.innerHTML+=out;
	},
	getCrystalNotes_onFailure:function(transport){
		$("xtalnotes_body").innerHTML='';
		if(404===transport.status){
			CrystalSelection.writeCrystalNotesForm();
			return;
		}
		alert("Could not retrieve notes for this crystal.");
	},
	writeCrystalNotesForm:function(){
		let out='';
		if(!canUpdate){ return; }
		out+='<div id="crystalnotesadd">';
		out+='<textarea id="crystalnotesaddtext" placeholder="Add a note..."></textarea>';
		out+='<br/><input id="crystalnotesaddbutton" type="button" value="Save note" />';
		out+='</div><hr/>';
		$("xtalnotes_body").innerHTML+=out;
		window.setTimeout(function(){ $("crystalnotesaddbutton").observe("click",CrystalSelection.addCrystalNote); },50);
	},

	addCrystalNote: function(){
		let cn=$("crystalnotesaddtext");
		if(cn.hasClassName("updating")){ return false; }
		let parentId=0;
		CrystalSelection.crystals.each(function(c){
			if(c.numberindrop==CrystalSelection.currentCrystalNumber){
				parentId=c.id;
			}
		});
		if(0===parentId){
			alert("Could not determine crystal for note");
			return false;
		}
		let noteText=cn.value.trim();
		if(""===noteText){
			alert("Cannot add an empty note");
			return false;
		}
		cn.addClassName("updating");
		new Ajax.Request('/api/note',{
			method:'post',
			parameters:{
				csrfToken:csrfToken,
				parentid:parentId,
				text:noteText,
			},
			onSuccess:CrystalSelection.addCrystalNote_onSuccess,
			onFailure:CrystalSelection.addCrystalNote_onFailure,
		});
	},
	addCrystalNote_onSuccess: function(){
		CrystalSelection.refresh();
	},
	addCrystalNote_onFailure: function(transport){
		$("crystalnotesaddtext").removeClassName("updating");
		let err=transport.responseText;
		if(transport.responseJSON && transport.responseJSON.error){
			err=transport.responseJSON.error;
		}
		alert("Could not add note. The server said:\n\n"+err);
	},

	deleteCurrentCrystal:function(){
		let sp=$("selectpanel");
		if(!confirm("Really delete crystal "+sp.currentCrystal.numberindrop+"?")){ return false; }
		$("xtaldelete").up("div").addClassName("updating");
		Crystal.delete(sp.currentCrystal, CrystalSelection.deleteCurrentCrystal_afterSuccess, null);
	},
	deleteCurrentCrystal_afterSuccess:function(response){
		$("selectpanel").currentCrystal=null;
		$$(".xs_crosshair").each(function(xh){
			if(xh.dataset.crystalid===1*response['deleted']){
				xh.remove();
			}
		});
		CrystalSelection.switchFrom();
		CrystalSelection.crystals=null;
		let img=DropViewer.getCurrentImage();
		img.crystals=null;
		CrystalSelection.switchTo();
	},
	
	addCrystal:function(evt){
		let crystalNumber=0;
		CrystalSelection.crystals.each(function(c){
			crystalNumber=Math.max(crystalNumber,c.numberindrop);
		});
		crystalNumber++;
		let img=DropViewer.getCurrentImage();
		let ci=$("currentimage");
		let offset=ci.cumulativeOffset();
		let scale=ci.getHeight() / img.pixelheight;
		let crystalX=Math.round((evt.clientX-offset.left)/scale);
		let crystalY=Math.round((evt.clientY-offset.top)/scale);
		let crystalName=DropViewer.getWellNameForImage(img)+'d'+img.dropnumber+'c'+crystalNumber;
		new Ajax.Request('/api/crystal',{
			method:'post',
			parameters:{
				csrfToken:csrfToken,
				numberindrop:crystalNumber,
				name:crystalName,
				pixelx:crystalX,
				pixely:crystalY,
				welldropid:img.welldropid,
				dropimageid:img.id,
				projectid:img.projectid
			},
			onSuccess:CrystalSelection.addCrystal_onSuccess,
			onFailure:CrystalSelection.addCrystal_onFailure
		});
	},
	addCrystal_onSuccess: function(transport){
		if(!transport.responseJSON|| !transport.responseJSON.created){
			return CrystalSelection.addCrystal_onFailure(transport);
		}
		CrystalSelection.crystals.push(transport.responseJSON.created);
		CrystalSelection.currentCrystalNumber=transport.responseJSON.numberindrop;
		CrystalSelection.refresh();
	},
	addCrystal_onFailure: function(transport){
		let err=transport.responseText;
		if(transport.responseJSON && transport.responseJSON.error){
			err=transport.responseJSON.error;
		}
		alert("Could not add crystal. The server said:\n\n"+err);
	},
	
	
	/* Movie controls when in crystals view */
	goToFirst:function(){
		DropViewer.goToFirst();
		CrystalSelection.switchFrom();
		CrystalSelection.switchTo();
	},
	goToLast:function(){
		DropViewer.goToLast();
		CrystalSelection.switchFrom();
		CrystalSelection.switchTo();
	},
	goToNext:function(){
		DropViewer.goToNext();
		CrystalSelection.switchFrom();
		CrystalSelection.switchTo();
	},
	goToPrevious:function(){
		DropViewer.goToPrevious();
		CrystalSelection.switchFrom();
		CrystalSelection.switchTo();
	},

	playOrStop:function(){
		if(DropViewer.movieInterval){
			CrystalSelection.stopMovie();
		} else {
			CrystalSelection.startMovie();
		}
	},

	startMovie:function(){
		$$(".xs_crosshair").each(function(ch){ 
			if("xs_crosshair"===ch.id){
				ch.style.display="none";
			} else {
				ch.remove(); 
			} 
		});
		DropViewer.startMovie();
	},

	stopMovie:function(){
		DropViewer.stopMovie();
		CrystalSelection.switchFrom();
		CrystalSelection.switchTo();
	},
	
	
	goUpOneRow:function(){
		DropViewer.goUpOneRow();
		CrystalSelection.switchFrom();
		CrystalSelection.switchTo();
	},
	goDownOneRow:function(){
		DropViewer.goDownOneRow();
		CrystalSelection.switchFrom();
		CrystalSelection.switchTo();
	},

	
};



/**********************************************************************************************************
 * MEASUREMENT 
 **********************************************************************************************************/

window.Measure={

	suppressed:true,
	
	init:function(){
		let ip=$("imagepane");
		ip.observe("mouseover", Measure.mouseOver);
		ip.observe("mousemove", Measure.mouseMove);
		ip.observe("mouseout",  Measure.mouseOut);
	},
			
	stop:function(){
		let ip=$("imagepane");
		ip.stopObserving("mouseover", Measure.mouseOver);
		ip.stopObserving("mousemove", Measure.mouseMove);
		ip.stopObserving("mouseout",  Measure.mouseOut);
	},
			
	setScale:function(){
		let m=$("dv_measure");
		let ci=$("currentimage");
		if(!ci){ return false; }
		let micronsHigh=ci.dataset.micronshigh;
		let pixelHeight=ci.getHeight();
		let screenMicronsPerPixel=micronsHigh/pixelHeight;
		m.style.height=Math.round(2000/screenMicronsPerPixel)+"px";
		m.style.width =Math.round(2000/screenMicronsPerPixel)+"px";
		if(screenMicronsPerPixel<=0){
			m.addClassName("dv_nomeasure");
		} else {
			m.removeClassName("dv_nomeasure");
		}
		if(screenMicronsPerPixel>3){
			m.addClassName("dv_smallmeasure");
		} else {
			m.removeClassName("dv_smallmeasure");
		}
	},
	mouseOver:function(){
		$("dv_measure").style.display="block";
		Measure.mouseMove();

	},
	mouseMove:function(evt){
		let m=$("dv_measure");
		if(!evt){ return false; }
		let absX=evt.pointer().x;
		let absY=evt.pointer().y;
		let imgPos=$("currentimage").getBoundingClientRect();
		let h=m.getHeight();
		let x=absX-imgPos.left;
		let y=absY-imgPos.top;
		m.style.left=(x)-(h/2)+"px";
		m.style.top=(y)-(h/2)+"px";
	},
	mouseOut:function(){
		$("dv_measure").style.display="none";
	},
	
	toggleVisibility:function(){
		if(Measure.suppressed){
			Measure.suppressed=false;
			Measure.init();
			Measure.mouseOver();
		} else {
			Measure.suppressed=true;
			Measure.mouseOut();
			Measure.stop();
		}
	},
	
};


/**********************************************************************************************************
 * DROP NAVIGATION
 **********************************************************************************************************/

window.DropNav={

	isOpen:false,
		
	onclose:function(){
		DropNav.isOpen=false;
		$("modalBox").removeClassName("dropviewer");
		document.stopObserving("keyup",DropNav.navigateToDropOnEnterKey);
		return true;
	},

	open:function(){
		let pt=DropViewer.plateType;
		if(!pt){ alert("No plate type info, cannot show navigation"); return false; }
		DropNav.isOpen=true;
		ui.modalBox({
			title:"Drop navigation",
			content:"Just a moment...",
			onclose:DropNav.onclose
		});
		window.setTimeout(DropNav._doOpen,50);
	},
	_doOpen:function(){
		let mb=$("modalBox");
		mb.addClassName("dropviewer");
		let pt=DropViewer.plateType;
		// let dropCells=[];
		let dm=pt.dropmapping.split(",");
		// let dmRows=dm.length;
		// let dmCols=dm[0].length;

		let out='<table id="dropnav"><tr><th>Drop</th><td>&nbsp;</td>';
		for(let i=1;i<=pt.cols;i++){
			out+="<th>"+i+"</th>";
		}
		out+="</tr>";
		let inBottomHalf=0;
		for(let r=1;r<=pt.rows;r++){
			if(r>pt.rows/2){
				inBottomHalf=1;
			}
			out+="<tr>";
			if(r<=pt.subs){
				out+='<th class="droppicker droppickerempty" id="droppicker'+r+'" data-dropnumber="'+r+'" onclick="DropNav.selectDropNumber('+r+')">'+r+'</th>';
			} else {
				out+="<th>&nbsp;</th>";
			}
			out+="<th>"+DropViewer.rowLabels[r]+"</th>";
			for(let c=1;c<=pt.cols;c++){
				out+='<td id="nav_'+r+'_'+c+'" data-in-bottom-half="'+inBottomHalf+'" data-row="'+r+'" data-col="'+c+'" class="dropnav_well empty" style="">';
				out+="</td>";
			}
			out+="</tr>";
		}
		out+="</tr></table>";
		
		mb.down(".boxbody").innerHTML=out;
		mb.down(".boxbody").down("tr").select("td,th").each(function(cell){
			cell.style.width=cell.getHeight()+"px";
		});
		let counter=0;
		DropViewer.images.each(function(img){
			let td=$("nav_"+img.row+"_"+img.col);
			td.removeClassName("empty");
			
			let dv=document.createElement("div");
			dv.id="dropnav_"+img.row+"_"+img.col+"_"+img.dropnumber;
			dv.className="dropnav_dropimage dropnav_drop"+img.dropnumber;
			dv.style.display="none";
			dv.dataset.row=img.row;
			dv.dataset.col=img.col;
			dv.dataset.dropnumber=img.dropnumber;
			dv.dataset.dropindex=counter+"";
			dv.dataset[img.lighttype.toLowerCase()+'image']=img.thumbnailurl;
			let score=$("score"+img.latestcrystalscoreid);
			if(score){
				dv.style.backgroundColor=score.style.borderColor;
			}
			let im=document.createElement("img");
			im.src=img.thumbnailurl;
			dv.appendChild(im);
			let lbl=document.createElement("div");
			lbl.addClassName("dropnav_comparelabel");
			lbl.innerHTML="Drop "+img.dropnumber;
			dv.appendChild(lbl);
			td.appendChild(dv);
			dv.observe("click",function(){ DropNav.goToDrop(dv); });
			dv.observe("mouseover",function(){ DropNav.startCompareDrops(dv); });
			dv.observe("mouseout",function(){ DropNav.stopCompareDrops(dv); });
			$("droppicker"+img.dropnumber).removeClassName("droppickerempty");
			counter++;
		});
		$$(".droppickerempty").each(function(dp) {
			dp.onclick=null;
			dp.style.cursor="not-allowed";
		});
		let currentImage=DropViewer.images[DropViewer.currentIndex];
		$("dropnav_"+currentImage.row+"_"+currentImage.col+"_"+currentImage.dropnumber).up("td").addClassName("dropnav_wellcurrent");
		document.observe("keyup", DropNav.navigateToDropOnEnterKey);
		window.setTimeout(function(){
			DropNav.selectDropNumber(currentImage.dropnumber);
			DropNav.setOtherLightTypeImages();
		},50);
	},
	
	otherImagingSession:null,
	setOtherLightTypeImages:function(){
		if(!DropNav.otherImagingSession || !DropNav.otherImagingSession.dropimages){
			window.setTimeout(DropNav.getOtherImagingSession,50);
			return;
		}
		if(0===DropNav.otherImagingSession.dropimages.length){ return false; }
		let lighttype=DropNav.otherImagingSession.dropimages[0].lighttype.toLowerCase();
		DropNav.otherImagingSession.dropimages.each(function(di){
			$("dropnav_"+di.row+"_"+di.col+"_"+di.dropnumber).dataset[lighttype+"image"]=di.thumbnailurl;
		});
		let th=$("dropnav").down("th");
		th.style.background="url(/images/icons/"+skin.bodyicontheme+"/switchlight.png) no-repeat center";
		th.style.cursor="pointer";
		th.innerHTML="";
		th.title="Switch between visible and UV images";
		th.onclick=DropNav.toggleLightTypeImages;
	},
	
	/**
	 * Examine the images currently shown in the drop nav pane, and retrieve an
	 * imaging session with the opposite light type (visible vs UV).
	 */
	getOtherImagingSession:function(){
		let dropNavBox=$("dropnav");
		let otherLightType=false;
		let dv=dropNavBox.down("[data-visibleimage]");
		let currentLightType="Visible";
		if(dv){
			otherLightType="UV";
		} else {
			dv=dropNavBox.down("[data-uvimage]");
			if(dv){
				otherLightType="Visible";
				currentLightType="UV";
			}
		}
		if(!otherLightType){ return false; }
		new Ajax.Request("/api/imagingsession/plateid/"+data['plateid']+"/lighttype/"+otherLightType+"?sortby=imageddatetime&sortdescending=yes",{
			method:"get",
			onSuccess: DropNav.getOtherImagingSession_onSuccess,
			onFailure: DropNav.getOtherImagingSession_onFailure,
		});
	},
	getOtherImagingSession_onSuccess:function(transport){
		//take the first one in transport.responseJSON.rows - was sorted latest first
		DropNav.otherImagingSession=transport.responseJSON.rows[0];
		DropNav.getOtherImagingSessionImages();
	},
	getOtherImagingSession_onFailure:function(){
		//nothing to do, just swallow it
	},
	getOtherImagingSessionImages:function(){
		new Ajax.Request("/api/imagingsession/"+DropNav.otherImagingSession.id+"/dropimage?pagesize=5000",{
			method:"get",
			onSuccess: DropNav.getOtherImagingSessionImages_onSuccess,
			onFailure: DropNav.getOtherImagingSessionImages_onFailure,
		});
	},
	getOtherImagingSessionImages_onSuccess:function(transport){
		DropNav.otherImagingSession.dropimages=transport.responseJSON.rows;
		DropNav.setOtherLightTypeImages();
	},
	getOtherImagingSessionImages_onFailure:function(){
		//nothing to do, just swallow it
	},

	currentLightType:"Visible",
	toggleLightTypeImages:function(){
		DropNav.currentLightType= ("Visible"===DropNav.currentLightType) ? "UV" : "Visible";
		let lt=DropNav.currentLightType.toLowerCase();
		$$(".dropnav_dropimage").each(function(dv){
			dv.style.border="green";
			dv.down("img").src=dv.dataset[lt+"image"];
		});
	},
	
	startCompareDrops:function(dropDiv){
		let mb=$("modalBox");
		mb.absolutize();
		let tr=dropDiv.up("tr");
		let td=dropDiv.up("td");
		let compareDiv=document.createElement("div");
		compareDiv.id="dropnav_compare";
		compareDiv.innerHTML=""+dropDiv.up("td").innerHTML.replace(' id="dropnav_', ' id="dropnavcompare_');
		compareDiv.setStyle({ position:"absolute" });
		mb.down(".boxbody").appendChild(compareDiv);
		compareDiv.select(".dropnav_dropimage").each(function(di){
			di.setStyle({ display:"inline-block", height:"auto", width:"auto" });
		});
		Element.clonePosition(compareDiv,dropDiv, { setWidth:false, setHeight:false, offsetLeft:-1000, offsetTop:100 });
		let w1=compareDiv.getWidth();
		let h1=compareDiv.getHeight();
		let w2=tr.getWidth();
		let h2=tr.getHeight();
		let w3=dropDiv.getWidth();
		let l2=tr.cumulativeOffset().left;
		let l3=dropDiv.up("td").previous().cumulativeOffset().left;

		let offsetLeft=5+l3+w3-l2;
		offsetLeft=Math.min(offsetLeft, w2-w1);

		//Filthy user-agent detect, but this doesn't need to be reliable. Worst-case, it pushes the drop images a bit
		//further away from the well. Without doing this, Chrome puts them over the well, causing the mouseout to fire
		//and an annoying flicker.
		let isChrome = /Chrome/.test(navigator.userAgent) && /Google Inc/.test(navigator.vendor);

		if(1===parseInt(td.dataset.inBottomHalf)){
			if(isChrome){
				Element.clonePosition(compareDiv,dropDiv.up("tr"), { setWidth:false, setHeight:false, offsetLeft:offsetLeft, offsetTop:-(h1+h2+h2+15) });
			} else {
				Element.clonePosition(compareDiv,dropDiv.up("tr"), { setWidth:false, setHeight:false, offsetLeft:offsetLeft, offsetTop:-(compareDiv.getHeight()+3) });
			}
		} else {
			Element.clonePosition(compareDiv,dropDiv.up("tr"), { setWidth:false, setHeight:false, offsetLeft:offsetLeft, offsetTop:dropDiv.getHeight()+9 });
		}

	},
	
	stopCompareDrops:function(){
		let compare=$("dropnav_compare");
		if(compare){ compare.remove(); }
	},
	
	selectDropNumber:function(dropNumber){
		//IE is "special" so we have to handle it "specially". Greeeeaaaaat....
		let isIE=/Trident\/|MSIE/.test(window.navigator.userAgent);
		let mb=$("modalBox");
		mb.select("div.dropnav_dropimage").each(function(dv){
			dv.style.display="none";
		});
		mb.select("div.dropnav_drop"+dropNumber).each(function(dv){
			dv.setStyle({ display:"block" });
			if(isIE){
				let wrapper=dv.up();
				let bw=parseInt(wrapper.currentStyle.borderTopWidth);
				dv.style.height=wrapper.getHeight()-(2*bw)+"px";
				dv.style.width=wrapper.getWidth()-(2*bw)+"px";
			}
		});
		mb.select("th.droppicker").each(function(dv){
			dv.removeClassName("droppickercurrent");
		});
		$("droppicker"+dropNumber).addClassName("droppickercurrent");
	},
	
	navigateToDropOnEnterKey:function(evt){
		let mb=$("modalBox");
		if(13!==evt.keyCode || !mb){ return; }
		let selectedDrop=mb.down("div.dropnav_wellcurrent");
		if(selectedDrop){ DropNav.goToDrop(selectedDrop); }
	},
	
	goToDrop:function(dropElement){
		switchPanels($("viewbutton_dv"));
		DropViewer.setCurrentImage(dropElement.dataset.dropindex);
		ui.closeModalBox();
	},
	
	goToFirst:function(){
		DropNav.goToIndexInNavWindow(0);
	},
	goToPrevious:function(byDrop){
		if(byDrop || KeyboardShortcuts.ctrlPressed){
			DropNav.moveInNavWindow(0,0,-1);
		} else {
			DropNav.goLeftOneColumn();
		}
	},
	goToNext:function(byDrop){
		if(byDrop || KeyboardShortcuts.ctrlPressed){
			DropNav.moveInNavWindow(0,0,1);
		} else {
			DropNav.goRightOneColumn();
		}
	},
	goToLast:function(){
		DropNav.goToIndexInNavWindow(DropViewer.images.length-1);
	},

	goLeftOneColumn:function(){
		DropNav.moveInNavWindow(0,-1,0);
	},
	goRightOneColumn:function(){
		DropNav.moveInNavWindow(0,1,0);
	},
	goUpOneRow:function(){
		DropNav.moveInNavWindow(-1,0,0);
	},
	goDownOneRow:function(){
		DropNav.moveInNavWindow(1,0,0);
	},

	moveInNavWindow:function(rowDelta, colDelta, dropDelta){
		//what's highlighted now?
		let mb=$("modalBox");
		let highlightedWell=mb.down("td.dropnav_wellcurrent");
		let highlightedDrop=mb.down("th.droppickercurrent");
		if(!highlightedWell || !highlightedDrop){ return false; }
		let oldRow=parseInt(highlightedWell.dataset.row);
		let oldCol=parseInt(highlightedWell.dataset.col);
		let oldDrop=parseInt(highlightedDrop.dataset.dropnumber);
		let toHighlight=null;
		let newDrop=null;
		if(0!==dropDelta){
			let oldIndex=parseInt($("dropnav_"+oldRow+"_"+oldCol+"_"+oldDrop).dataset.dropindex);
			DropNav.goToIndexInNavWindow(oldIndex+dropDelta);
		} else if(0!==rowDelta){
			let newRow=oldRow+rowDelta;
			toHighlight=$("dropnav_"+newRow+"_"+oldCol+"_"+oldDrop);
			if(!toHighlight){ return; }
			DropNav.goToIndexInNavWindow(toHighlight.dataset.dropindex);
		} else if(-1===colDelta){
			let prev=highlightedWell.previous("td.dropnav_well");
			if(!prev){
				let prevRow=highlightedWell.up("tr").previous("tr");
				if(prevRow){ prev=prevRow.select("td.dropnav_well").last(); }
			}
			if(prev){ 
				newDrop=prev.down(".dropnav_drop"+oldDrop);
				if(newDrop){
					DropNav.goToIndexInNavWindow(newDrop.dataset.dropindex); 
				}
			}
		} else if(1===colDelta){
			let next=highlightedWell.next("td.dropnav_well");
			if(!next){
				let nextRow=highlightedWell.up("tr").next("tr");
				if(nextRow){ next=nextRow.down("td.dropnav_well"); }
			}
			if(next){ 
				newDrop=next.down(".dropnav_drop"+oldDrop);
				if(newDrop){
					DropNav.goToIndexInNavWindow(newDrop.dataset.dropindex); 
				}
			}
		}
	},

	goToIndexInNavWindow:function(index){
		let img=DropViewer.images[index];
		if(!img){ return; }
		let dropDiv=$("dropnav_"+img.row+"_"+img.col+"_"+img.dropnumber);
		if(!dropDiv){ return; }
		$$(".dropnav_wellcurrent").each(function(elem){ elem.removeClassName("dropnav_wellcurrent")});
		dropDiv.up("td").addClassName("dropnav_wellcurrent");
		dropDiv.addClassName("dropnav_wellcurrent");
		DropNav.selectDropNumber(dropDiv.dataset.dropnumber);
		DropNav.stopCompareDrops();
		DropNav.startCompareDrops(dropDiv);
	},
	
};

/**********************************************************************************************************
 * SMALL SCREEN
 * If the screen width is below the threshold (see ui.js:isSmallScreen), render a stripped-down viewer.
 **********************************************************************************************************/
window.SmallScreenDropViewer={

		init:function(){
			DropViewer.getPlateType();
			$("imagepane").remove();
			$("controls").remove();
			let h1=document.querySelector("header h1");
			h1.innerHTML=h1.innerHTML.replace("Trial drop viewer: ","").replace(" drop "," ");
			$("container").setStyle({
				"position":"absolute", "top":0, "bottom":0, "right":0, "left":0,
				"height":window.height+"px"
			});
			let content=$("content");
			content.setStyle({
				"position":"absolute", "bottom":0, "right":0, "left":0,
				"top":$(document.body).down("header").getHeight()+"px",
				"height":window.height+"px"
			});
			content.addClassName("smallscreenviewer");
			content.innerHTML='<div id="imagecache" style="position:absolute;left:-1000px"></div><div id="ssimagepane">Getting images...</div><div id="sscontrols"></div>';
			new Ajax.Request('/api/imagingsession/'+data['id']+'/dropimage?pagesize=10000',{
				method:'get',
				onSuccess:SmallScreenDropViewer.init_onSuccess,
				onFailure:SmallScreenDropViewer.init_onFailure
			});
		},
		init_onFailure:function(){
			$("ssimagepane").innerHTML="Could not get images.";
		},
		init_onSuccess:function(transport){
			$("imagecache").innerHTML='';
			DropViewer.images=transport.responseJSON.rows;
			DropViewer.maxIndex=DropViewer.images.length-1;
			let startImage=0;
			if(document.location.hash){
				let dropLabel=document.location.hash.substring(1);
				let row=DropViewer.rowLabels.indexOf(dropLabel.substr(0,1));
				let parts=dropLabel.substring(1).split(".");
				let col=parseInt(parts[0]);
				let drop=parts[1];
				if(false!==drop.indexOf("c")){
					let subparts=drop.split("c");
					drop=subparts[0];
				}
				let numImages=DropViewer.images.length;
				for(let i=0;i<numImages;i++){
					let img=DropViewer.images[i];
					if(row===parseInt(img.row) && col===parseInt(img.col) && drop===parseInt(img.dropnumber)){
						startImage=i;
						DropViewer.currentIndex=i;
						break;
					}
				}
			}
			SmallScreenDropViewer.renderControls();
			SmallScreenDropViewer.setCurrentImage(startImage);
		},
		setCurrentImage:function(index){
			DropViewer.currentIndex=index;
			let img=DropViewer.images[index];
			if($("cached"+img.id)){
				$("ssimagepane").innerHTML='<img src="/api/dropimagethumb/'+img.id+'" alt=""/>';
			} else {
				$("ssimagepane").innerHTML='Getting image...';
				window.setTimeout(function(){
					$("ssimagepane").innerHTML='<img src="/api/dropimagethumb/'+img.id+'" alt=""/>';
				},50);
			}
			//update the page title
			DropViewer.setDropNameInTitle(img);
			//update the controls
			//cache adjacent images
			//			transport.responseJSON.rows.each(function(img){
			//				$("imagecache").innerHTML+='<img src="/api/dropimagethumb/'+img.id+'"/>';
			//			});
			SmallScreenDropViewer.updateControls();
		},

		renderControls:function(){
			let ssImagePane=$("ssimagepane");
			let ssControls=$("sscontrols");
			let content=$("content");
			if(ssImagePane.getHeight()<100 || !ssImagePane.down("img")){
				window.setTimeout(SmallScreenDropViewer.renderControls, 100);
				return false;
			}
			ssControls.innerHTML="";
			//This mess is because Unicode has a pile of poo symbol but no 4-way arrow symbol. We overlay a N-S and E-W arrow.
			ssControls.innerHTML+='<div id="ssnav" class="sscontrol" onclick="SmallScreenDropViewer.showNav();return false">'+
				'<div style="position:absolute;top:0;left:0;right:0;bottom:0;display:table-cell;text-align:center">&#x2194;</div>'+
				'<div style="position:absolute;top:0;left:0;right:0;bottom:0;display:table-cell;text-align:center">&#x2195;</div>'+
			'</div>';

			//disable scoring control
			//ssControls.innerHTML+='<div id="ssscore" class="sscontrol" href="#" onclick="SmallScreenDropViewer.showScoring();return false">&#x2714;</div>';
			
			ssControls.innerHTML+='<div id="ssprev"  class="sscontrol" onclick="SmallScreenDropViewer.prevDrop();return false">&lt;</div>';
			ssControls.innerHTML+='<div id="ssnext"  class="sscontrol" onclick="SmallScreenDropViewer.nextDrop();return false">&gt;</div>';
			ssControls.innerHTML+='<div id="ssfirst" class="sscontrol" onclick="SmallScreenDropViewer.firstDrop();return false">&#x23EA;</div>';
			ssControls.innerHTML+='<div id="sslast"  class="sscontrol" onclick="SmallScreenDropViewer.lastDrop();return false">&#x23E9;</div>';
			ssControls.innerHTML+='<div id="ssplay"  class="sscontrol" onclick="SmallScreenDropViewer.playStop();return false">&#x25B6;</div>';
			if(content.getHeight()>content.getWidth()){
				//Portrait
				//We assume that images are landscape, and that the phone is tall enough to fit both image and min-height controls,
				//so we don't need to scale the image down. Instead, bring the top of the controls up to meet the image.
				ssImagePane.down("img").style.width="100%";
				ssImagePane.style.height=ssImagePane.down("img").getHeight()+"px";
				ssControls.style.height=(content.getHeight()-ssImagePane.getHeight())+"px";
				ssControls.style.width="100%";
			} else {
				//Landscape
				//Assuming that the image is landscape, and that the controls must be a minimum width to be usable, we may end up with:
				ssControls.style.height="100%";
				ssControls.style.width="4cm";
				ssControls.style.right="0";
				ssControls.style.left="auto";
				ssImagePane.style.right="4cm";
				ssImagePane.down("img").style.height="20%";
				ssImagePane.down("img").style.width="auto";
				let widthFraction=ssImagePane.down("img").getWidth()/ssImagePane.getWidth();
				let heightFraction=ssImagePane.down("img").getHeight()/ssImagePane.getHeight();
				if(widthFraction<heightFraction){
					//width-limited. Set image height 100%.
					ssImagePane.down("img").style.height="100%";
					ssImagePane.down("img").style.width="auto";
				} else {
					//height-limited. Set image width 100%.
					ssImagePane.down("img").style.height="auto";
					ssImagePane.down("img").style.width="100%";
				}
				ssImagePane.style.width=ssImagePane.down("img").getWidth()+"px";
				ssControls.style.width=(content.getWidth()-ssImagePane.getWidth())+"px";
			}
			SmallScreenDropViewer.updateControls();
		},
		
		updateControls:function(){
			
			//if playing movie, cache image 5 out and cull first in cache (on start movie, cache first five)
			//else wipe cache and cache adjacent in X,Y,drop axes
			if(!SmallScreenDropViewer.isPlaying){
				$$(".sscontrol").each(function(elem){
					elem.style.opacity="1";
				});
				if(0===DropViewer.currentIndex){
					$("ssprev").style.opacity="0.4";
					$("ssfirst").style.opacity="0.4";
				} else if(DropViewer.maxIndex===DropViewer.currentIndex){
					$("ssnext").style.opacity="0.4";
					$("sslast").style.opacity="0.4";
					$("ssplay").style.opacity="0.4";
				}
			}
		},
		
		movieInterval:null,
		movieDelay:0.8, //seconds
		isPlaying:false,
		playStop:function(){
			if(SmallScreenDropViewer.isPlaying){
				SmallScreenDropViewer.stopMovie();
			} else {
				SmallScreenDropViewer.startMovie();
			}
		},

		startMovie:function(){
			let play=$("ssplay");
			if(SmallScreenDropViewer.isPlaying || DropViewer.currentIndex===DropViewer.maxIndex){ return false; }
			play.innerHTML="&#x25FC;";
			play.style.color="#f00";
			$$(".sscontrol").each(function(elem){
				elem.style.opacity="0.4";
			});
			play.style.opacity="1";
			SmallScreenDropViewer.isPlaying=true;
			//grey out all controls except Stop
			//cache next five images
			SmallScreenDropViewer.movieInterval=window.setInterval(SmallScreenDropViewer.advanceMovie, SmallScreenDropViewer.movieDelay*1000);
		},
		
		stopMovie:function(){
			let play=$("ssplay");
			play.innerHTML="&#x25B6;";
			play.style.color="#0f0";
			$$(".sscontrol").each(function(elem){
				elem.style.opacity="1";
			});
			window.clearInterval(SmallScreenDropViewer.movieInterval);
			SmallScreenDropViewer.isPlaying=false;
			SmallScreenDropViewer.updateControls();

		},

		advanceMovie:function(){
			DropViewer.currentIndex++;
			SmallScreenDropViewer.setCurrentImage(DropViewer.currentIndex);
			if(DropViewer.currentIndex===DropViewer.maxIndex){
				SmallScreenDropViewer.stopMovie();
			}
		},

		nextDrop:function(){
			if(SmallScreenDropViewer.isPlaying || DropViewer.currentIndex===DropViewer.maxIndex){ return false; }
			SmallScreenDropViewer.setCurrentImage(DropViewer.currentIndex+1);
		},
		
		prevDrop:function(){
			if(SmallScreenDropViewer.isPlaying || DropViewer.currentIndex===0){ return false; }
			SmallScreenDropViewer.setCurrentImage(DropViewer.currentIndex-1);
		},

		firstDrop:function(){
			if(SmallScreenDropViewer.isPlaying || DropViewer.currentIndex===0){ return false; }
			SmallScreenDropViewer.setCurrentImage(0);
		},

		lastDrop:function(){
			if(SmallScreenDropViewer.isPlaying || DropViewer.currentIndex===DropViewer.maxIndex){ return false; }
			SmallScreenDropViewer.setCurrentImage(DropViewer.maxIndex);
		},

		showNav:function(row,col){
			ui.closeModalBox();
			let pt=DropViewer.plateType;
			let items=[];
			if(!row){
				//choosing row
				for(let i=1;i<=pt.rows;i++){
					items.push({
						"label":DropViewer.rowLabels[i],
						"row":i,
						"col":null,
						"hasImages":false,
						"bestScoreIndex":null
					});
				}
				DropViewer.images.each(function(img){
					items[img.row-1].hasImages=true;
				});
			} else if(!col){
				//choosing col
				let rowLabel=DropViewer.rowLabels[row];
				for(let i=1;i<=pt.cols;i++){
					items.push({
						"label":rowLabel+i,
						"row":row,
						"col":i,
						"hasImages":false,
						"bestScoreIndex":null
					});
				}
				DropViewer.images.each(function(img){
					items[img.col-1].hasImages=true;
				});
			} else {
				//choosing drop.
				let colLabel=(Plate.getWellName(row,col,1).split("."))[0];
				for(let i=1;i<=pt.subs;i++){
					items.push({
						"label":colLabel+"."+i,
						"row":row,
						"col":col,
						"drop":i,
						"hasImages":false,
						"bestScoreIndex":null,
						"dropImageIndex":null
					});
				}
				for(let i=0;i<=DropViewer.maxIndex;i++){
					let img=DropViewer.images[i];
					if(!img){ break; }
					if(row===parseInt(img.row) && col===parseInt(img.col)){
						items[img.dropnumber-1].hasImages=true;
						items[img.dropnumber-1].dropImageIndex=i;
					}
				}
				let imageCount=0;
				let imageIndex=-1;
				items.each(function(i){
					if(i.hasImages){ 
						imageCount++;
						imageIndex=i.dropImageIndex;
					}
				});
				if(1===imageCount && -1!==imageIndex){
					SmallScreenDropViewer.setCurrentImage(imageIndex);
					return false;
				}
			}

			//Later, maybe show best score in row/col?
			
			//render the nav buttons
			ui.modalBox({
				title:'Choose row, column, drop',
				content:''
			});
			let bb=$("modalBox").down(".boxbody");
			bb.style.padding="0";
			bb.style.borderRadius="0";
			bb.style.backgroundColor="black";
			bb.style.overflow="hidden";
			let isPortrait=(bb.getHeight()>bb.getWidth());
			let halfDimension="height";
			let otherDimension="width";
			let halfAxis="top";
			let otherAxis="left";
			let numButtons=items.length;
			let halfButtons=Math.max(numButtons/2);
			let xPercent=100/halfButtons;
			let yPercent=50;
			let breakpoint=halfButtons;
			if(isPortrait){
				halfDimension="width";
				otherDimension="height";
				halfAxis="left";
				otherAxis="top";
				breakpoint=2;
				yPercent=100/halfButtons;
				xPercent=50;
			}
			if(numButtons<5){
				if(isPortrait){
					breakpoint=1;
					xPercent=100;
					yPercent=100/numButtons;
				} else {
					breakpoint=numButtons;
					yPercent=100;
					xPercent=100/numButtons;
				}

			} 
			let xCount=0;
			let yCount=0;
			items.each(function(i){
				let btn=document.createElement("div");
				btn.className="ssnavitem";
				btn.style.top=(yCount*yPercent)+"%";
				btn.style.left=(xCount*xPercent)+"%";
				btn.style.height=(yPercent-2)+"%";
				btn.style.width=(xPercent-2)+"%";
				if(i.dropImageIndex){
					btn.onclick=function(){ 
						SmallScreenDropViewer.setCurrentImage(i.dropImageIndex);
						ui.closeModalBox();
					};
				} else if(i.hasImages){
					btn.onclick=function(){ SmallScreenDropViewer.showNav(i.row, i.col); };
				} else {
					btn.style.opacity="0.4";
				}
				btn.innerHTML=i.label;
				bb.appendChild(btn);
				xCount++;
				if(xCount>=breakpoint){ xCount=0; yCount++; }
			});
		}
		
};


/**********************************************************************************************************
 * OTHER FUNCTIONS
 **********************************************************************************************************/

function togglePageConfig(){
	
	let handedness=UserConfig.get("dropviewer_handedness","right");
	let bb=ui.modalBox({
		title:'Configure the drop viewer',
		content:''
	});
	
	let f=bb.form({});
	f.setStyle({
		marginTop:0,
		position:'absolute',
		top:"1%",
		left:"1%",
		width:'65%'
	});
	
	let handednessDiv=document.createElement("div");
	handednessDiv.id="dv_config_handedness";
	
	let lh=document.createElement("div");
	lh.id="dv_config_lh";
	let leftImage=document.createElement("img");
	leftImage.src='/images/icons/dropviewer/controlsleft.png';
	leftImage.alt='Controls on the left';
	leftImage.style.height="auto";
	lh.appendChild(leftImage);
	lh.onclick=function(){ setHandedness("left"); };
	if("left"===handedness){ lh.style.borderColor='#666'; }
	handednessDiv.appendChild(lh);
	
	let rh=document.createElement("div");
	rh.id="dv_config_rh";
	let rightImage=document.createElement("img");
	rightImage.src='/images/icons/dropviewer/controlsright.png';
	rightImage.alt='Controls on the right';
	rightImage.style.height="auto";
	rh.appendChild(rightImage);
	rh.onclick=function(){ setHandedness("right"); };
	if("right"===handedness){ rh.style.borderColor='#666'; }
	handednessDiv.appendChild(rh);

	let lbl1=f.formField({
		"label":"Layout - controls on left or right",
		"content":""
	});
	lbl1.appendChild(handednessDiv);

	let lbl2=f.formField({
		label:"Minimum score for Shift-left/right navigation",
		content:""
	});
	let scoresDiv=document.createElement("div");
	scoresDiv.id="dv_configscorethresholdwrapper";
	scoresDiv.style.maxWidth="70%";
	scoresDiv.style.cssFloat="right";
	lbl2.appendChild(scoresDiv);
	writeScoreThresholdButtons(scoresDiv);
	lbl2.innerHTML+='<div style="clear:both;max-height:0.25em;">&nbsp;</div>';

	let f2=bb.form({});
	f2.setStyle({
		position:'absolute',
		top:"1%",
		right:"1%",
		width:'32%'
	});
	
	//SELECT for panel to show first
	let opts=[
		{ value:'DropViewer', label:'Scoring' },
		{ value:'TimeCourse', label:'Imaging history' },
		{ value:'DropInfo', label:'Drop info' },
		{ value:'CrystalSelection', label:'Crystals' },
	];
	f2.dropdown({
		label:'Show this at start',
		name:'dropviewer_startMode',
		options:opts,
		value:UserConfig.get('dropviewer_startMode'),
		apiUrl:'/api/userconfig/dropviewer_startMode'
	});
	
	let lbl=f2.formField({label:'Items to show on info panel', content:'&nbsp;',name:'bogus' });
	lbl.addClassName("radiohead");

	Object.keys(DropInfo.infoPanelItems).each(function(k){
		let label=k.split("_show_");
		if(label[1]){ label=label[1].replace("_"," ")}
		f2.checkbox({ label:label, name:k, handler:setInfoPanelItemVisibility, checked:1*DropInfo.infoPanelItems[k]});
	})	
	
}
function writeScoreThresholdButtons(wrapper,currentThreshold){
	let scoringSystemId=DropViewer.scores[0].crystalscoringsystemid;
	wrapper.innerHTML="";
	if(undefined===currentThreshold){
		currentThreshold=1*UserConfig.get("dropviewer_scorethreshold_ss"+scoringSystemId,0);
	}
	DropViewer.scores.each(function(sc){
		let out='<div data-scoreindex="'+sc.scoreindex+'" onclick="setShiftArrowScoreThreshold('+scoringSystemId+',this.dataset.scoreindex)" style="';
		out+='display:inline-block;width:14em;margin:0.25em;pointer:cursor;color:black;textShadow:1px 1px 1px white;';
		out+='padding=left:0.5em;text-align:left;border:4px solid #'+sc.color+';';
		if((1*sc.scoreindex)>=(1*currentThreshold)){
			out+='background-color:#'+sc.color+';';
			out+='font-weight:bold;';
			if((1*sc.scoreindex)===(1*currentThreshold)){
				out+='border:4px solid black;';
			}
		}
		out+='">'+sc.label+'</div>';
		wrapper.innerHTML+=out;
	});
}

/**
 * Click handler for info panel item checkboxes.
 * Not intended to be called directly, instead passed as a handler when
 * writing the buttons in the config box
 * @param inp The hidden input associated with the checkbox image
 * @returns
 */
function setInfoPanelItemVisibility(inp){
	let lbl=inp.up("label");
	lbl.addClassName("updating");
	DropInfo.infoPanelItems[inp.name]=parseInt(inp.value);
	UserConfig.set(inp.name, inp.value);
	window.setTimeout(function(){ 
		DropInfo.refresh();
		lbl.removeClassName("updating");
	}, 500);
}

function setHandedness(hand){
	if("left"===hand){
		$(document.body).addClassName("lefthanded");
		UserConfig.set("dropviewer_handedness","left");
		if($("dv_config_handedness")){
			$("dv_config_lh").style.borderColor="#666";
			$("dv_config_rh").style.borderColor="transparent";
		}
	} else {
		$(document.body).removeClassName("lefthanded");
		UserConfig.set("dropviewer_handedness","right");
		if($("dv_config_handedness")){
			$("dv_config_lh").style.borderColor="transparent";
			$("dv_config_rh").style.borderColor="#666";
		}	
	}	
}

function setShiftArrowScoreThreshold(scoringSystemId,scoreIndex){
	if(isNaN(scoreIndex)){ scoreIndex=0; }
	scoreIndex=parseInt(scoreIndex);
	writeScoreThresholdButtons($("dv_configscorethresholdwrapper"),scoreIndex);
	UserConfig.set("dropviewer_scorethreshold_ss"+scoringSystemId, scoreIndex);
}

let currentPanel;
function switchPanels(button){
	let id=button.id;
	let newObj;
	if(id==="viewbutton_dv"){ newObj=DropViewer; }
	if(id==="viewbutton_tc"){ newObj=TimeCourse; }
	if(id==="viewbutton_di"){ newObj=DropInfo; }
	if(id==="viewbutton_xs"){ newObj=CrystalSelection; }
	if(newObj===currentPanel){ return false; }
	currentPanel.switchFrom();
	newObj.switchTo();
	currentPanel=newObj;
}
function setCurrentModeButton(button){
	$$(".viewbutton").each(function(vb){
		vb.removeClassName("current");
		vb.down("img").src=vb.down("img").src.replace("_current.png",".png");
	});
	button.addClassName("current");
	button.down("img").src=button.down("img").src.replace(".png","_current.png");
}
